/* HTML Library
 * Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * Author:  Jeff Heflin
 * Contact: Jeff Heflin (heflin@cs.umd.edu)
 *
 */

package HTML;

import java.lang.*;
import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

/** A tokenizer that given an input stream, parses it into HTML tokens. */
public class HTML_Tokenizer extends Object {
    
  public static final int HT_EOF = 0;
  public static final int HT_TEXT = 1;
  public static final int HT_SGMLTAG = 2;
  public static final int HT_COMMENT = 3;
  public static final int HT_SGMLDECL = 4;

  protected HTML_Token token;
  protected PushbackInputStream instream;

  public HTML_Tokenizer(InputStream inStream) throws IOException {

    instream = new PushbackInputStream(new BufferedInputStream(inStream));

  }

  /** Returns the next complete token from the input stream. A complete
    token is either surrounded by '<' and '>' or is text that is 
    preceded and followed by these characters. */
  /* We used to allow quotes to hide the tag delimeter characters, but
     if a single quote is omitted then determining where it went wrong
     is difficult. */
  /* It would be nice if we could somehow signal an unmatched quote. */
  public int nextToken() {

    int type = HT_TEXT;                           // type of token
    StringBuffer strbuf;
    char c;
    int b;
    boolean finished = false;
    boolean quoted = false;

    strbuf = new StringBuffer();

    try {
      b=instream.read();
      if (b < 0)                   // return null if end of input stream
	return HT_EOF;
      while (b >= 0 && finished == false) {
	c = (char) b;
	quoted = false;        // TEST: don't allow quotes to hide < or >
	if (quoted != true) {        // quotes can hide special characters
	  if (c == '<')
	    if (strbuf.length() == 0) {   // start of a new tag
	      type = HT_SGMLTAG;
	      strbuf.append(c);
	    }
	    else {   
	      if (type != HT_COMMENT) {  
		// end of text token or improperly ended tag
		finished = true;
		instream.unread(b);       // pushback character
	      }
	      else {                     // enclosed in a comment
		strbuf.append(c);
	      }
	    }
	  else if (c == '>') {
	    strbuf.append(c);
	    if (type == HT_SGMLTAG || type == HT_SGMLDECL)  // end of tag token
	      finished = true;
	    else if (type == HT_COMMENT) {    // must end with '-->'
	      int len = strbuf.length();
	      if (strbuf.charAt(len - 3)=='-' && strbuf.charAt(len - 2)=='-')
		finished = true;
	    }
	  }
	  else if (c == '"') {
	    if (type != HT_TEXT)   // quotes w/in a tag have special meaning
	      quoted = true;
	    strbuf.append(c);
	  }
	  else {                    // ordinary character, add to token
	    strbuf.append(c);

	    /* determine if this is an SGML declaration or comment */
	    if (strbuf.length() == 2 && type == HT_SGMLTAG) {
	      if (strbuf.charAt(1) == '!')
		type = HT_SGMLDECL;
	    }
	    else if (strbuf.length() == 4 && type == HT_SGMLDECL) {
	      if (strbuf.charAt(2) == '-' && strbuf.charAt(3) == '-')
		type = HT_COMMENT;
	    }
	  }
	} else {                      // quoted == true
	  strbuf.append(c);
	  if (c == '"')               // closing quote
	    quoted = false;
	}
	if (finished == false)    // if not done, read the next character
	  b = instream.read();
      }
    } catch (IOException ex) {
      System.out.println("Error reading input stream");
    }

    if (strbuf.length() == 0)
      type = HT_EOF;
    else {                          // create the token 
      if (type != HT_SGMLTAG)
	token = new HTML_Token(type, strbuf.toString());
      else
	token = SGML_Tag.parseTag(strbuf.toString());
    }
    return type;
  }
  
  /** Returns the type of the next token (HT_SGMLTAG, HT_COMMENT, HT_TEXT)
    and places the result in token. If this is an SGML tag, the token is
    parsed into attribute value pairs. */
  // not currently used
  /*
  public int nextToken() {

    int type = HT_TEXT;                // default is text
    // note: should eventually find out why it doesn't work without
    //   this default. 

    String tstr;
    StringBuffer ctxt;         // holds comment text

    tstr = _nextToken();
    if (tstr == null)
      type = HT_EOF;
    else {
      if (tstr.startsWith("<")) {
        if (tstr.startsWith("<!--")) {    // comment
          type = HT_COMMENT;
	  // the following code makes sure that comments end with "-->".
	  // Without it, a '>' that appears within a comment will terminate
	  // the comment 
	  ctxt = new StringBuffer(tstr);
	  while (tstr != null && tstr.endsWith("-->") == false) {  
	    tstr = _nextToken();
	    ctxt.append(tstr);
	  }
	  tstr = ctxt.toString();
	} 
        else if (tstr.startsWith("<!"))   // declaration
          type = HT_SGMLDECL;
        else {
          type = HT_SGMLTAG;
        }
      }
      else {                  // text
        type = HT_TEXT;
      }
      if (type != HT_SGMLTAG)
        token = new HTML_Token(type, tstr);
      else
        token = SGML_Tag.parseTag(tstr);
    }
    return type;
  }
  */


  /** Returns the last last token read by the tokenizer */
  public HTML_Token getLastToken() {

    return token;
  } 

  /** Returns text corresponding to token */
  public String toString()
  {
    return token.toString();
  }

}

