/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in HTML_Tokenizer.java
 * and the file License.html for details.
 */

package HTML;

import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;

/** Class that represents a single SGML tag.
 */
public class SGML_Tag extends HTML_Token {

  /** The element name of the tag. */
  private String name;

  /** The tag's attributes, stored as SGML_Attrib objects. */
  private Vector attlist;

  /** True if this is an XML-style tag with no content or closing tag,
    i.e., it has a "/" at the end. */
  private boolean xmlEmptyTag = false; 

  /** Given a string that represents an SGML tag, returns the equivalent
    object of type SGML tag. */
  public static SGML_Tag parseTag(String tstr) {

    TagTokenizer ttkns;
    SGML_Tag tag;
    String aname = null;
    String value = null;
    String nextToke;
    
    ttkns = new TagTokenizer(tstr);
    tag = new SGML_Tag(ttkns.nextToken());      // name tag with first token
      
    if (ttkns.hasMoreTokens())
      aname = ttkns.nextToken();
    while (ttkns.hasMoreTokens()) {
      nextToke = ttkns.nextToken();             // read whitespace token
      if (nextToke.equals("=")) {
        value = ttkns.nextToken();
	// System.out.println("Attrib=" + aname + ", Value=" + value);
	tag.addAttribute(aname, value);
	aname = ttkns.nextToken();
      }
      else {
        value = aname;
        aname = null;
        // System.out.println("Attrib=" + aname + ", Value=" + value);
       	tag.addAttribute(aname, value);
       	aname = nextToke;
      }
    }

    /* if a token is left over... */
    if (aname != null) {

      /* if it is "/", consider this an XML-style empty tag */
      if (aname.equals("/"))
	tag.setXmlEmptyTag(true);

      /* otherwise, it is an SGML value without a matching name */
      else
	tag.addAttribute(null, aname);
    }
    return tag;
  }

  
  /** Construct a new tag with no name or attributes. To be valid SGML,
   at least a name must be set. */
  public SGML_Tag() {

    type = HTML_Token.TT_SGMLTAG;
    name = "";
    attlist = new Vector();
    xmlEmptyTag = false;
  }


  /** Given a string, creates an SGML tag with that name*/
  public SGML_Tag(String str) {

    type = HTML_Token.TT_SGMLTAG;
    if (str != null)
      name = str;
    else
      name = "";
    attlist = new Vector();
    xmlEmptyTag = false;
  }


  /** Adds an attribute with the given value to the tag. */
  public void addAttribute(String aname, String value) {
    
    SGML_Attrib sattrib = new SGML_Attrib(aname, value);

    attlist.addElement(sattrib);
  }


  /** Returns the name of the attribute that has the given value. */
  public String findValName(String value) {

    SGML_Attrib attrib = null;
    
    attrib = getAttByValue(value);
    if (attrib == null)
      return null;
    else
      return attrib.getName();
  }


  /** Given an SGML attribute name returns the value or null if there
    is none. Since SGML is case-insensitive, the name is returned in
    a cannonical format, i.e., upper case. */
  public String getAttribute(String aname) {
  
    SGML_Attrib attrib = null;
    
    attrib = getAttByName(aname);
    if (attrib == null)
      return null;
    else
      return attrib.getValue();
  }

  /** Returns the name of the tag in cannonical format. Since SGML is
    case-insensitive, this returns the name in upper case. If we change
    to XML, this should return the name without changing the case. */
  public String getName() {

    return name.toUpperCase();
  }

  /** Returns the textual representation of the SGML tag. Overrides the
    HTML_Token method. */
  public final String getText() {

    return toString();
  }

  /** Returns true if this tag is an XML style tag without
    content or a corresponding end tag. */
  public boolean isXmlEmptyTag() {
    
    return xmlEmptyTag;
  }

  
  /** Sets the name of the attribute with the given value. Used to
     assign default names to attributes that don't have them. */
  public void setValName(String value, String aname) {

    SGML_Attrib attrib = null;
    
    attrib = getAttByValue(value);
    if (attrib != null)
      attrib.setName(aname);
  }

  /** Sets the value of the given attribute as specified. */
  public void setValue(String aname, String value) {

    SGML_Attrib attrib = null;
    
    attrib = getAttByName(aname);
    if (attrib != null)
      attrib.setValue(value);
  }

  /** Sets the status of whether this tag is an XML style tag without
    content or a corresponding end tag. If true, the string representation
    of this tag will contain a "/" before the closing angle bracket. 
    */
  public void setXmlEmptyTag(boolean xmlEmptyTag) {

    this.xmlEmptyTag = xmlEmptyTag;
  }

  /** Returns the SGML string equivalent of the tag. */
  public String toString() {

    StringBuffer tstr;
    Enumeration attributes = attlist.elements();

    tstr = new StringBuffer("<" + name);
    while (attributes.hasMoreElements())
      tstr.append(" " + attributes.nextElement().toString());
    if (isXmlEmptyTag())
      tstr.append(" /");    
    tstr.append(">");
    return tstr.toString();
  }



  /* NOTE: The functions getAttributeValue() and getAttributeName() are only
     used in SHOEUpgrage.Tran. Consider removing them. */
  public String getAttributeValue(int index) {
  
    SGML_Attrib attrib;
    attrib = (SGML_Attrib)attlist.elementAt(index);

    if (attrib == null)
      return null;
    else
      return attrib.getValue();
  }

  /** Returns the name of the attribute at the position determined by
    index.
    */
  public String getAttributeName(int index) {
    SGML_Attrib attrib;
    attrib = (SGML_Attrib)attlist.elementAt(index);  
 
    if (attrib == null)
      return null;
    else
      return attrib.getName();
  }


  /* **************** Private Methods ******************* */

  /** Given an SGML attribute name, returns the attribute's object. */
  private SGML_Attrib getAttByName(String aname) {
    
    int size;
    int index;
    boolean found = false;
    SGML_Attrib attrib = null;
    
    size = attlist.size();
    index = 0;
    while (index < size && found == false) {
      attrib = (SGML_Attrib)attlist.elementAt(index);
      if (attrib.getName().equalsIgnoreCase(aname))
        found = true;
      else
        index = index + 1;
    }
    if (found == true)
      return attrib;
    else
      return null;
  }

  /** Given a particular attribute value, returns an attribute object
    that has that value. */  
  private SGML_Attrib getAttByValue(String value) {

    
    int size;
    int index;
    boolean found = false;
    SGML_Attrib attrib = null;
    
    size = attlist.size();
    index = 0;
    while (index < size && found == false) {
      attrib = (SGML_Attrib)attlist.elementAt(index);
      if (attrib.getValue().equalsIgnoreCase(value))
        found = true;
      else
        index = index + 1;
    }
    if (found == true)
      return attrib;
    else
      return null;
  }

}

/* ******************************************************************** */

/** A SGML attribute. It consists of a name and a value. */
class SGML_Attrib extends Object {
	
  String name;
  String value;
  
  public SGML_Attrib() {
  
    name = "";
    value = "";
  }
  
  public SGML_Attrib(String name, String value) {
    
    if (name != null)
      this.name = name;
    else
      this.name = "";           // attributes can't have null names
    this.value = value;
  }
 
  public String getName() {
  
    return name;
  }

  public String getValue() {
  
    return value;
  }

  public void setName(String newname) {
  
    name = newname;
  }

  public void setValue(String newval) {
  
    value = newval;
  }

  public String toString() {

    String str;

    if (name == null)
      str = value;
    else
      str = name + "=" + '"' + value + '"';
    return str;
  }

}


/* ******************************************************************** */

class TagTokenizer extends StringTokenizer {

  boolean pushedBack = false;
  String pushToken = null;

  public TagTokenizer(String str) {

    /* if the last argument is false, then the delimiters won't be 
       considered tokens. This may be what we want!!!! If we don't
       consider whitespace as tokens, we have a problem with
       quoted strings because the enclosed whitespace will be ignored. */
    super(str, "<> \n\t\f\r\"\'=", true);
    readWhiteSpace();

    // StringTokenizer ttkns;
    // String ignoreChars = "<> \n\t\f\r";
    // String quoteDelims = "\"\'";
    // String normDelims = ignoreChars + quoteDelims + "=";
  } 
  
  void pushBack(String str) {
    
      pushToken = str;
      pushedBack = true;
  }

  public String nextToken() {
  
    String toke = null;
    StringBuffer strb;
    boolean endQuote;
    char quoteChar;

    /* ignore white space and tag delimiters */
    readWhiteSpace();

    if (pushedBack == true) {
      toke = pushToken;
      pushedBack = false;
    }
    else {
      if (super.hasMoreTokens() == false)
	return null;
      else
       	toke = super.nextToken();
    }

    /* check for quoted strings */
    if (toke.equals("\"") || toke.equals("\'")) {
      quoteChar = toke.charAt(0);
      endQuote = false;
      strb = new StringBuffer();
      while (super.hasMoreTokens() && endQuote == false) {
        toke = super.nextToken();
        if (toke.charAt(0) == quoteChar)
          endQuote = true;
        else
          strb.append(toke);
      }
      toke = strb.toString();
    }
    /* remove trailing whitespace */
    readWhiteSpace();
    return toke;
  }

  /* Reads past all whitespace characters. Also ignores the tag delimiters
    "<" and ">"  */     
  void readWhiteSpace() {

    boolean wspace = true;
    String toke;
    char c;                         // first character of a token

    if (pushedBack == false) {
      while (super.hasMoreTokens() && wspace == true) {
      	toke = super.nextToken();
	c = toke.charAt(0);
	if (Character.isSpace(c) || c=='<' || c=='>') {
	}
	/*
      	if (toke.equals("<") || toke.equals(">") || toke.equals(" ") ||
    	      toke.equals("\n") || toke.equals("\f") ||
	          toke.equals("\r") || toke.equals("\t")) {
      	}
	*/
      	else {
      	  wspace = false;
      	  pushBack(toke);
      	}
      }
    }
  }
}


/* ******************************************************************** */
