/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/* Change History:
   09/27/01, JDH: Changed validate() to check that the specified name is
                     actually a category
   11/01/00, JDH: Used accessor methods to get fields in Def_Category
   10/26/00, JDH: Added getForText(), setName(), and setForText(). 
                  Made name and for_text private                  
*/

package SHOE;

import java.util.*;
import java.io.*;
import HTML.*;

/** This class is used to describe and manipulate a SHOE Category. */
public class Category extends Declaration {

  private String name;
  private String for_text;
  byte usage;
  Unique_Elt elmt_id;              // id of the category
  Unique_Elt constId;              // id of a constant in the FOR field
  SHOE_Obj partOf;                 // object containing this category 

  /** Create a new category claim given the prefixed name of the category
    and the key of the subject. */
  public Category(String name, String for_text) {
		
    this.name = name;
    this.for_text = for_text;
    this.usage = U_VAL;
  }
	
  /** Create a category from its SGML tag representation. */
  public Category(SGML_Tag sgmltag) {
      
    String utext;
    
    name = sgmltag.getAttribute("NAME");
    for_text = sgmltag.getAttribute("FOR");
    usage = U_VAL;
    utext = sgmltag.getAttribute("USAGE");
    if (utext != null && utext.equals("VAR"))
      usage = U_VAR;
  }

  /** Returns an identifier for the Category. Note this identifier
    is NOT guaranteed to be unique. */
  public String getKey() {
  
    return name;
  }

  /** Returns the key of the instance that is classified under the category. */
  public String getForText() {

    return for_text;
  }

  /** Returns the prefixed name of the category. */
  public String getName() {

    return name;
  }

  /** Returns the name of the category as it appears in its originating
    ontology. */
  public String getBaseName() {

    if (elmt_id == null) {
      System.out.println("No elmt_id for category " + name); 
      return null;
    }
    else
      return elmt_id.getName();    
  }

  /** Returns a unique identifier for the category. Additionally, this
    identifier indicates which ontology the category originates from.  */
  public Unique_Elt getElmtId() {
    
    return elmt_id;
  }

 /** Returns a string that formats the category as a predicate, i.e.,
   is-a(subject, catname). */
  public String getPredicateText() {

    String str;
    Instance inst;

    inst = (Instance)partOf;
    str = "is-a(";
    /*    if (for_text == null)
      str = str + "me";
      */
    if (for_text.equals(inst.getKey()))
      str = str + "me";
    else 
      str = str + "\"" + for_text + "\"";
    str = str + ", \"" + name + "\")";
    return str;
  }
	
  /** Returns a phrase suitable for describing the claim given that
    its subject is already known. (i.e., is-a catname) */
  public String getTextBySubject(OntManager ontman) {

    String str = "";
    Instance inst;
    Def_Category dcat=(Def_Category)ontman.getElement(elmt_id);

    if (dcat == null)
      str = "ERROR: ";
    inst = (Instance)partOf;
    str = str + "is-a ";

    if (dcat == null || dcat.getPhrase() == null)
      str = str + name;
    else 	 
      str = str + dcat.getPhrase();
    return str;
  }    


  /** Returns true when the category declaration is for a variable. */
  public boolean isVariable() {

    if (usage == U_VAR)
      return true;
    else
      return false;
  }

  /** Sets the subject of the category claim. */
  public void setForText(String forText) {

    this.for_text = forText;
  }

  /** Sets the prefixed name of the category referred to in this claim. */
  public void setName(String name) {

    this.name = name;
  }

  /////////////////////////////////


  public Vector readSubTags(HTML_Tokenizer intokens) {

    Vector errStrings = null;

    errStrings = new Vector();
    return errStrings;
  }


  /** Given an SGML tag representing a SHOE category, determines if
    it contains the required fields and returns a boolean indicating
    if there are any errors. Specific error messages are appended to
    the errStrings vector. */
  public boolean checkFields(SGML_Tag sgmltag, Vector errStrings) {

    boolean       hasErrors = false;
    SHOEException perr;

    if (sgmltag.getAttribute("NAME") == null) {
      hasErrors = true;
      perr = new SHOEException(SHOEException.MISSING_FIELD,
			       sgmltag.toString(), "NAME");
      errStrings.addElement(perr.getMessage());
    }
    return hasErrors;
  }


  /** Checks the claim for errors and updates internal structures. */
  public Vector validate(SHOE_Obj parent, OntManager ontman) {
    /* is there a better way to access ontman without passing it? */

    Instance inst;
    SHOE_Obj obj = null;
    SHOEException perr;
    Vector errStrings = null;
    String forName;

    errStrings = new Vector();
    partOf = parent;

    /* create default when FOR attribute is null, empty or equals "ME".
       Categories in Ontology inferences cannot have default values. */
    if (parent instanceof Instance) {
      inst = (Instance)parent;
      if (for_text == null || for_text.length() == 0 || 
	  for_text.equalsIgnoreCase("ME"))
	for_text = inst.getKey();
    }

    /* resolve name of category */
    try {
      elmt_id = ontman.resolveName(name, parent);
    } catch (SHOEException ex) {
      errStrings.addElement(ex.getMessage());
      return errStrings;
    }
    //  System.out.println("Elmt Id = " + elmt_id);
    if (elmt_id == null) {
      System.out.println("Unexpected path in Category.validate()");
      perr = new SHOEException(SHOEException.NO_ELEMENT_DEF, name, "?");
      errStrings.addElement(perr.getMessage());
      return errStrings;
    }

    /* Next, check to see if the element really is a category */
   /** If the category declaration is part of an inference rule in an ontology,
      getElement() will get caught in an infinite loop 
      reloading the ontology. Therefore, we check if the category
      is defined in the current ontology, and if so, get its
      definition directly from the ontology, rather than accessing
      the ontology manager. */
    if (parent instanceof Instance)
      obj = ontman.getElement(elmt_id);    
    else if (parent instanceof Ontology) {
      Ontology ont = (Ontology)parent;
      if (ont.getId().equals(elmt_id.getOntId()) && 
	  ont.getVersion().equals(elmt_id.getOntVer()))
	  obj = ont.getElement(elmt_id.getName());
      else
	  obj = ontman.getElement(elmt_id);
    }

    if (obj == null) {
      perr = new SHOEException(SHOEException.NO_ELEMENT_DEF, name, "?");
      errStrings.addElement(perr.getMessage());
      return errStrings;
    }

    /* now that we have the object, check if its a Def-Category */
    if (!(obj instanceof Def_Category)) {
	perr = new SHOEException(SHOEException.NOT_CATEGORY,
				 elmt_id.getName(),
				 elmt_id.getOntId() + ", v." +
				 elmt_id.getOntVer());
	errStrings.addElement(perr.getMessage());
	return errStrings;
    }

    /* validate for_text if it's a constant */
    if (for_text.startsWith("!")) {
      forName = for_text.substring(1);
      try {
	constId = ontman.resolveName(forName, parent);
      } catch (SHOEException ex) {
	errStrings.addElement(ex.getMessage());
	return errStrings;
      }
    }

    /* Previous message said the following line isn't used. It also
       created infinite loops when validating categories used in
       inference rules. */
    /*
      Def_Category dcat;
      dcat = (Def_Category) ontman.getElement(elmt_id);
      */

    return errStrings;

  }

  /** Writes the tag and its contents (if any) in SGML format to the
    given print stream. */
  public void printSGML(PrintStream out) {
  
    out.print(this);
  }

  /** Returns the string representation of the tag. Any contents
   or end tags are not included. */
  public String toString() {
  
    StringBuffer str;
    
    str = new StringBuffer("<CATEGORY NAME=\"" + name + "\"");
    if (usage == U_VAR)
      str.append(" USAGE=\"VAR\"");
    if (for_text != null)
      str.append(" FOR=\"" + for_text + "\"");
    str.append(">");
    return str.toString();
  }
  
}

