/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/* Change history:
   10/26/00, JDH: Used accessor methods for Category fields
   */

package SHOE;

import java.util.*;

/** A supplemental class that will organize an instance in such a way
  that its claims can be efficiently retrieved by the subject of the
  claim. Here subject means the object that a category claim describes
  or an argument that a relation claim refers to. */
public class ClaimSubjectIndexer {

  Hashtable subjects;
  int       indexArity;

  /** Constructs a new ClaimSubjectIndexer using the category claims
    and arguments of relations up to the specified indexArity. */
  public ClaimSubjectIndexer(Instance inst, int indexArity) {

    Enumeration   enum;
    SHOE_Obj      obj;
    String        key;
    ClaimSubject  claimSub;
    Relation      rel;
    int           i;
    RelArg        arg;
    Category      cat;

    this.indexArity = indexArity;
    subjects = new Hashtable();

    enum = inst.getDeclarations();          // loop through claims
    while (enum.hasMoreElements()) {         
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Relation) {
	rel = (Relation)obj;
	for (i=1; i <= indexArity; i++) {   
	  arg = rel.getArg(i);               // indexing starts with 1
	  if (arg != null)
	    key = arg.getValue();
	  else
	    key = "Error";
	  claimSub = (ClaimSubject)subjects.get(key);
	  if (claimSub == null) {
	    claimSub = new ClaimSubject(key, indexArity);
	    claimSub.addRelation(rel, i);
	    subjects.put(key, claimSub);
	  }
	  else
	    claimSub.addRelation(rel, i);
	}
      }
      else if (obj instanceof Category) {
	cat = (Category)obj;
	key = cat.getForText();
	claimSub = (ClaimSubject)subjects.get(key);
	  if (claimSub == null) {
	    claimSub = new ClaimSubject(key, indexArity);
	    claimSub.addCategory(cat);
	    subjects.put(key, claimSub);
	  }
	  else
	    claimSub.addCategory(cat);
      }
      else
	System.out.println("Internal error in ClaimSubjectIndexer()");
    }
  }

  /** Returns the set of Categories for the subject with the given key. */
  public Vector getSubjectCategories(String key) {

    ClaimSubject claimSub;

    claimSub = (ClaimSubject)subjects.get(key);
    if (claimSub != null)
      return claimSub.getCategoryClaims();
    else
      return new Vector();
  }

  /** Returns the set of Relations in which the given subject key
    appears in the given argument position. */
  public Vector getSubjectRelations(String key, int argPos) {

    ClaimSubject claimSub;

    claimSub = (ClaimSubject)subjects.get(key);
    if (claimSub != null)
      return claimSub.getRelationClaims(argPos);
    else
      return new Vector();
  }

  /** Returns the set of claim subjects that have been indexed for this
    instance. A claim subject consists of a subject key and the set
    of all claims that refer to that subject. */
  public Enumeration getClaimSubjects() {

    return subjects.elements();
  }

  /** Returns the set of subject keys that have been indexed for this
    instance. */
  public Enumeration getClaimSubjectKeys() {

    Vector      keyVec;
    Enumeration subEnum;
    ClaimSubject claimSub;

    keyVec = new Vector();
    subEnum = subjects.elements();
    while (subEnum.hasMoreElements()) {
      claimSub = (ClaimSubject)subEnum.nextElement();
      keyVec.addElement(claimSub.getKey());
    }
    return keyVec.elements();
  }

}

class ClaimSubject {

  String   key;
  int      indexArity;
  Vector   catClaims;
  Vector[] relClaims;

  public ClaimSubject(String key, int indexArity) {

    int i;
    
    this.key = key;
    this.indexArity = indexArity;
    catClaims = new Vector();
    relClaims = new Vector[indexArity];
    for (i=0; i < indexArity; i++)
      relClaims[i] = new Vector();
  }

  public void addCategory(Category cat) {

    /* note that since this class is only used internally, we assume
       that it is correctly used and don't waste time checking if
       the claim is being indexed under the correct subject. */
    catClaims.addElement(cat);
  }

  public void addRelation(Relation rel, int argPos) {

    /* note that since this class is only used internally, we assume
       that it is correctly used and don't waste time checking if
       the claim is being indexed under the correct subject. */
    if (argPos <= indexArity)
      relClaims[argPos - 1].addElement(rel);
  }

  public String getKey() {

    return key;
  }

  public Vector getCategoryClaims() {

    return catClaims;
  }

  public Vector getRelationClaims(int argPos) {

    if (argPos <= indexArity)
      return relClaims[argPos - 1];
    else
      return new Vector();
  }
}
