/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/* Change History:
   11/03/00, JDH: Use accessor methods to get fields in RelArg.
*/

package SHOE;

import java.util.*;
import java.io.*;
import HTML.*;
import java.lang.System;

/** A class that describes a comparison clause for an inference rule. A
  comparison consists of an operator and two operands (also called
  arguments. */
public class Comparison extends Declaration {
  /* It might be better to make this a subclass of SHOE_Obj, since comparisons
     can only occur in inferences and are not technically declarations */

  Vector relargs; 
  int opType;
  //  public String name; 

  /** Create a new comparison clause given the type of the operator and
    an array of two operands. */
  public Comparison(int opType, RelArg[] args) {
		
    this.opType = opType;
    this.relargs = new Vector();
    if (args.length >= 2) {
      relargs.addElement(args[0]);
      relargs.addElement(args[1]);
    }
  }
	
  /** Create a new comparsion clause from its SGML tag representation. */
  public Comparison(SGML_Tag sgmltag) {
      
    String opName;

    relargs = new Vector();

    opType = C_UNDEFINED;
    opName = sgmltag.getAttribute("OP");
    if (opName != null) {
      if (opName.equals("equal"))
	opType = C_EQUAL;
      else if (opName.equals("notEqual"))
	opType = C_NOT_EQUAL;
      else if (opName.equals("greaterThan"))
	opType = C_GREATER;
      else if (opName.equals("greaterThanOrEqual"))
	opType = C_GREATER_OR_EQUAL;
      else if (opName.equals("lessThanOrEqual"))
	opType = C_LESS_OR_EQUAL;
      else if (opName.equals("lessThan"))
	opType = C_LESS;
    }
  }

  public String getKey() {
  
    return null;
  }

  /** Return the operator type of the comparison. This is either
    C_EQUAL, C_NOT_EQUAL, C_GREATER, C_GREATER_OR_EQUAL, C_LESS,
    or C_LESS_OR_EQUAL. */
  public int getOperator() {
  
    return opType;
  }

  public String getOperatorName() {

    String opName = null;

    switch (opType) {
    case C_EQUAL:
      opName = "equal";
      break;
    case C_NOT_EQUAL:
      opName = "notEqual";
      break;
    case C_GREATER:
      opName = "greaterThan";
      break;
    case C_GREATER_OR_EQUAL:
      opName =  "greaterThanOrEqual";
      break;
    case C_LESS:
      opName = "lessThanOrEqual";
      break;
    case C_LESS_OR_EQUAL:
      opName = "lessThan";
      break;
    default:
      opName = "UNDEFINED";
    }
    return opName;
  }

  /** Return the argument of the comparison with the given position number.
    The LHS is position 1 and the RHS is position 2. */
  public RelArg getArg(int pos) {

    int i;
    RelArg arg;

    if (relargs.size() >= 2) 
      ;	//error message
    else {
      for (i=0; i < relargs.size(); i++) {
	arg = (RelArg)relargs.elementAt(i);
	if (arg.getPosition() == pos)
	  return (RelArg)relargs.elementAt(i);
      }
    }
    return null;
  }

  /** Return an array of the operands of the comparison. */
  public RelArg[] getArgs() {

    int i;
    RelArg arglist[];

    arglist = new RelArg[relargs.size()];
    for (i=0; i < relargs.size(); i++)
      arglist[i] = (RelArg)relargs.elementAt(i);
    return arglist;
  }

  /** Return a predicate representation of the comparison. */
  public String getPredicateText() {

    String str;
    int numarg;
    RelArg arg;
    int blankargs = 0;

    str = getOperatorName() + "(";
    numarg = relargs.size();
    for (int i=0; i < numarg; i++) {
      arg = (RelArg)relargs.elementAt(i);
      if (arg.getPosition() + blankargs > i + 1) {
	blankargs = blankargs + 1;
	str = str + "me, ";
      }
      str = str + "\"" + arg.getValue() + "\"";
      if (i < numarg - 1) 
	str = str + ", ";
    }
    str = str + ")";
    return str;
  }

  /** Return a subject style representation of the comparison. */
  public String getTextBySubject(OntManager ontman) {

    String str = "";
    int numarg;
    RelArg arg;
    int blankargs = 0;

    numarg = relargs.size();
    for (int i=0; i < numarg; i++) {
      arg = (RelArg)relargs.elementAt(i);
      if (arg.getPosition() + blankargs > i + 1) {
	blankargs = blankargs + 1;
	str = str + "me " + getOperatorName() + " ";
      }
      str = str + "\"" + arg.getValue() + "\"";
      if (i < numarg - 1) 
	str = str + " " + getOperatorName() + " ";
    }
    return str;
  }

  /** Set the value of the argument at the selected position. */
  public void setArgValue(int pos, String value) {

    RelArg arg;

    arg = getArg(pos);
    arg.setValue(value);
  }
	
  public Vector readSubTags(HTML_Tokenizer intokens) {

    boolean endtag = false;
    int ttype;
    HTML_Token token;
    SGML_Tag sgmltag = null;
    String tagname;
    RelArg arg;
    Vector errStrings = null;

    errStrings = new Vector();
    do {
      ttype = intokens.nextToken();
      token = intokens.getLastToken();
      if (token instanceof SGML_Tag) {
        sgmltag = (SGML_Tag)token;
        tagname = sgmltag.getName();
        if (tagname.equals("/COMPARISON"))
          endtag = true;
        else if (tagname.equals("ARG")) {
          arg = new RelArg(sgmltag);
          relargs.addElement(arg);
        }
        else
          ;         // insert invalid tag error handling
      }
    } while (ttype != intokens.HT_EOF && endtag == false);
    return errStrings;
  }

  public void printSGML(PrintStream out) {
    
    Enumeration enum;
    RelArg arg;
    
    out.println(this);
    enum = relargs.elements();
    while (enum.hasMoreElements()) {          /* print arguments */
      arg = (RelArg)enum.nextElement();
      out.print("   ");
      arg.printSGML(out);
    }
    out.println("</COMPARISON>");
  }

  /** Returns the string representation of the tag. Any contents or end
    tags are not included.  */
  public String toString() {
  
    String str;
    
    str = "<COMPARISON OP=\"";
    str = str + getOperatorName(); 
    str = str + "\">";
    return str;
  }
  
}

