/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/*
  Change History:
    01/10/01, JDH: Added a makeCopy() method
                   Changed short_phrase to shortPhrase
    10/26/00, JDH: Used accessor methods for fields in Ontology,
                   Use_Ontology and Unique_Elt
 */

package SHOE;

import java.util.*;
import java.io.*;
import HTML.*;

/** This class implements facilities for defining an argument to a SHOE
   relation. */
public class Def_Arg extends SHOE_Obj {

  private int pos = 0;
  private String type = null;
  private String shortPhrase;
  private Unique_Elt typeId = null;         
     /* Unique element that references a Def_Category or Def_Type */

  /* constants */
  /** Constant for identifying an unknown type. */
  public static final int UNDEFINED = -1;

  /** Constant for identifying a numerical type. */
  public static final int NUMBER = 1;

  /** Constant for identifying a string type. */
  public static final int STRING = 2;

  /** Constant for identifying a date type. */
  public static final int DATE = 3;

  /** Constant for identifying a truth data type (i.e., YES/NO) */
  public static final int TRUTH = 4;

  /** Constant for identifying a type that is identified by a category. */
  public static final int CATEGORY = 5;


  /** Create a new argument definition given its position in the relation
    and type. */
  public Def_Arg(int pos, String type) {
		
    this.pos = pos;	
    this.type = type;
  }

  /** Create a new argument definition given its position in the relation,
    type, and a short phrase. */  
  public Def_Arg(int pos, String type, String shortPhrase) {

    this.pos = pos;
    this.type = type;
    this.shortPhrase = shortPhrase;
  }

  /** Create a new argument definition given its SGML tag representation. */
  public Def_Arg(SGML_Tag sgmltag) {

    String posstr;
    
    posstr = sgmltag.getAttribute("POS");
    if (posstr != null) {
      if (posstr.equals("FROM"))
        pos = 1;
      else if (posstr.equals("TO"))
        pos = 2;
      else {
	try {
	  pos = Integer.parseInt(posstr);
	} catch (NumberFormatException ex) {
	  pos = 0;
	}
      }
    }
    else {
      if (sgmltag.findValName("FROM").equals("")) 
        pos = 1;
      else if (sgmltag.findValName("TO").equals(""))
        pos = 2;
    }
    type = sgmltag.getAttribute("TYPE");
    shortPhrase = sgmltag.getAttribute("SHORT");
  }

  /** Return a string that distinguishes the argument from other
    arguments in the same relation. */
  public String getKey() {
  
    return Integer.toString(pos);

  }

  /** Return a short phrase that describes the argument. */
  public String getPhrase() {
   
    return shortPhrase;
  }


  /** Return the sequence number of the argument in the relation. */
  public int getPosition() {

    return pos;
  }

  /** Return the prefixed name of the argument's type. */
  public String getType() {
    
    return type;
  }

  /** Return the class of the argument's type. The class can be
    NUMBER, STRING, DATE, TRUTH or CATEGORY, each of which is
    represented by a constant. */   
  public int getTypeClass() {
    /* This is a quick fix. Should do prefix resolution on type name */
    int typeid;

    if (type.endsWith("NUMBER"))
      typeid = NUMBER;
    else if (type.endsWith("STRING"))
      typeid = STRING;
    else if (type.endsWith("DATE"))
      typeid = DATE;
    else if (type.endsWith("TRUTH"))
      typeid = TRUTH;
    else
      typeid = CATEGORY;
    return typeid;
  }

  /** If the argument is of the type of an object defined in an ontology,
    returns the unique identifier for that object, otherwise returns
    null. */
  public Unique_Elt getTypeId() {

    return typeId;
  }

  /** If the argument has a SHORT name, it is returned, otherwise the
     name of the type is returned */
  public String getLabel() {

    String label;

    if (shortPhrase != null)
      label = shortPhrase;
    else
      label = type;
    return label;
  }

  /** Returns a copy of the Def_Arg. */
  public Def_Arg makeCopy() {

    Def_Arg argCopy;

    argCopy = new Def_Arg(pos, type, shortPhrase);
    return argCopy;
  }

  public Vector readSubTags(HTML_Tokenizer intokens) {

    Vector errStrings = null;

    errStrings = new Vector();
    return errStrings;
  }

  /** Resolves the type of each argument to make sure it is the name of a
    Def_Category or Def_Constant and assigns the corresponding id. */
  public Vector validate(SHOE_Obj parent, OntManager ontman) {

    Vector errStrings;
    SHOE_Obj obj;
    SHOEException err;
    Ontology ont;

    errStrings = new Vector();

    if (type != null) {        // can't validate if there is no type 
      try {
	typeId = ontman.resolveName(type, parent);
      } catch (SHOEException ex) {
	errStrings.addElement(ex.getMessage());
	return errStrings;
      }

      //  System.out.println("Arg=" + this.toString() + ", TypeId=" + typeId);

      /* get the object representing the type. If the type is from the
	 current ontology then get it from the ontology. Otherwise, make
	 a request to the OntManager (note that the current ontology is not
	 yet stored in the OntManager). */
      ont = (Ontology)parent;
      if (ont.getId() == typeId.getOntId() &&
	  ont.getVersion() == typeId.getOntVer())
	obj = ont.getElement(typeId.getName());
      else 
	obj = ontman.getElement(typeId);
      if (obj instanceof Def_Category || obj instanceof Def_Type)
	;
      else {
	// System.out.println("Elmt_id = " + typeId);
	err = new SHOEException(SHOEException.INVALID_TYPE, type);
	errStrings.addElement(err.getMessage());
      }
    }
    return errStrings;
  }

  /** Given a type name, sets the argument definitions type to it. */
  public void setType(String data_type) {

    type = data_type;
  }

  public void printSGML(PrintStream out) {
  
    out.println(this);
  }

  /** Returns the string representation of the tag. Any contents
   or end tags are not included. */
  public String toString() {
  
    StringBuffer str;
    
    str = new StringBuffer("<DEF-ARG POS=\"" + pos + "\"");
    str.append(" TYPE=\"" + type + "\"");
    if (shortPhrase != null)
      str.append(" SHORT=\"" + shortPhrase + "\"");
    str.append(">");
    return str.toString();
  }
  
}
