/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/*
  Change History:
    01/10/01, JDH: Changed short_para to shortPhrase
    11/03/00, JDH: Made all fields private
    10/26/00, JDH: Used accessor methods for fields in Ontology and Unique_Elt
                   Added getParentNames() method
*/

package SHOE;

import java.util.*;
import java.io.*;
import HTML.*;

/** A object that represents a <DEF-CATEGORY> tag in a SHOE document. */
public class Def_Category extends SHOE_Obj {
	
  private String name;
  private Vector isa;   		 // vector of parent category list
  private String description;
  private String shortPhrase;
  private Vector parent_ids;            // vector of parent element ids
  private Vector child_ids;             // vector of child category ids

  /** Creates a new Def_Category object with the given name and set of
    direct super-categories. The super-categories are specified as
    Strings, where each is a SHOE prefixed name. */
  public Def_Category(String name, Vector isa) {

    this.name = name;
    this.isa = isa;
    this.description = null;
    this.shortPhrase = null;
    parent_ids = new Vector();
    child_ids = new Vector();
  }
  
  /** Creates a new Def_Category object with the given name, set of
    direct super-categories, description, and short phrase. */
  public Def_Category(String name, Vector isa, String description, String shortPhrase) {

    this.name = name;
    this.isa = isa;
    this.description = description;
    this.shortPhrase = shortPhrase;
    parent_ids = new Vector();
    child_ids = new Vector();
  }

  /** Creates a new Def_Category object using the given SGML tag. */
  public Def_Category(SGML_Tag sgmltag) {
 
     StringTokenizer tknzr;
     String isastr;

     name = sgmltag.getAttribute("NAME");
     description = sgmltag.getAttribute("DESCRIPTION");
     shortPhrase = sgmltag.getAttribute("SHORT");

     isastr = new String();
     isastr = sgmltag.getAttribute("ISA");
     isa = new Vector();
     if (isastr != null) {
	tknzr = new StringTokenizer(isastr);
	while (tknzr.hasMoreTokens())
	   isa.addElement(tknzr.nextToken());
     }

    parent_ids = new Vector();
    child_ids = new Vector();
  }

  /** Returns a description of the category. */
  public String getDescription() {

    return description;
  }


  public String getKey() {
  
    return name;
  }


  /** Returns the name of the category. */
  public String getName() {
    
    return name;
  }

  
  /** Returns an array of the prefixed names of the category's parents. */
  public String[] getParentNames() {

    String[] pNameArray = new String[isa.size()];
    isa.copyInto(pNameArray);
    return pNameArray;
  }

  /** Returns a short phrase that describes the category. This is
    the SHORT field in the DEF-CATEGORY tag. */
  public String getPhrase() {

    return shortPhrase;
  }


  /** Returns a label suitable for display the Category name to a user.
    If the category has a SHORT name, it is returned, otherwise the
    name of the category is returned */
  public String getLabel() {

    String label;

    if (shortPhrase != null)
      label = shortPhrase;
    else
      label = name;
    return label;
  }

  /** Returns the list of prefixed names of parent categories as a 
    whitespace separated string. */
  public String getIsaString() {

    String str = "";
    int cnt;
    int i;

    cnt = isa.size();
    for (i=0; i < cnt; i++)
      str = str + isa.elementAt(i) + " ";
    str.trim();
    return str;
  }

  /** Returns the unique element ids of each parent category, including
    those of categories defined in other ontologies. */
  public Enumeration getParentIds() {

    return parent_ids.elements();
  }

  /** Returns the unique element ids of each child category, including
    those of categories defined in other ontologies. */
  /* NOTE: We may also want to create a method that only returns the child
     categories with local names. */
  public Enumeration getChildIds() {

    return child_ids.elements();
  }

  public Vector readSubTags(HTML_Tokenizer intokens) {

    Vector errStrings = null;

    errStrings = new Vector();
    return errStrings;
  }

  /** Validate the Def_Category by determining the element ids for each
    of its parents. In the process we also add the current category
    as a child to each of those. Here, parent is the ontology that
    contains the Def-Category. */

  /* BUG: We cannot re-validate simply by calling this procedure
    again. The current element must be removed from the child lists
    of its parents, and then it's parent vector can be reset. */

  public Vector validate(SHOE_Obj parent, OntManager ontman) {

    Enumeration enum;
    String catname;
    Unique_Elt elmt_id = null;
    Vector errStrings;

    errStrings = new Vector();

    enum = isa.elements();
    while (enum.hasMoreElements()) {
      catname = (String)enum.nextElement();
      try {
	elmt_id = ontman.resolveName(catname, parent);
      } catch (SHOEException ex) {
	errStrings.addElement(ex.getMessage());
	return errStrings;
      }

      /* BUG: if category has previously been validated, then this will add
	 all of its parents in duplicate. */
      parent_ids.addElement(elmt_id);

      // add this category to a list of its parent's children
      if (elmt_id != null) {
	Def_Category parCat;
	Ontology ont;

	ont = (Ontology)parent;
	if (ont.getId().equals(elmt_id.getOntId()) &&
	    ont.getVersion().equals(elmt_id.getOntVer()))
	  parCat = (Def_Category)ont.getElement(elmt_id.getName());
	else
	  parCat = (Def_Category)ontman.getElement(elmt_id);

	/* BUG: if category has previously been validated, then this
	   will add it to its parents in duplicate. */
	if (parCat != null)
	  parCat.addChildCat(new Unique_Elt(ont.getId(), ont.getVersion(),
					    name));
      }

      // System.out.println("Cat: " + name + ", elmt_id = " + elmt_id);
    }
    return errStrings;
  }

  public void printSGML(PrintStream out) {
  
    out.println(this);
  }

  /** Returns the string representation of the tag. Any contents
   or end tags are not included. */
  public String toString() {
  
    String str;
    
    str = "<DEF-CATEGORY NAME=\"" + name + "\"";
    if (isa.isEmpty() == false)
      str = str + " ISA=\"" + getIsaString() + "\"";
    if (description != null)
      str = str + " DESCRIPTION=\"" + description + "\"";
    if (shortPhrase != null)
      str = str + " SHORT=\"" + shortPhrase + "\"";
    str = str + ">";
    return str;
  }

  /** This is a private function meant only for use by the validate()
    method. It should be called on a category object after that object
    has been made the parent of the object identified by elmt_id. */
  private void addChildCat(Unique_Elt elmt_id) {

    child_ids.addElement(elmt_id);
  }
}

