/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

package SHOE;

import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.*;
import HTML.*;

/** A class the represents a SHOE inference rule. The rule consists of
 an IF part and a THEN part, each of which may have multiple clauses. */
public class Def_Inference extends SHOE_Obj {

  private Vector infIf;
  private Vector infThen;
  private String description;

  /** Create a new inference rule whose IF and THEN parts will be set
    later. */
  public Def_Inference()  {

    this.description = null;
    infIf = new Vector();
    infThen = new Vector();
  }
  
  /** Create a new inference rule given a description. The IF and THEN
    parts will be set later. */
  public Def_Inference(String description) {

    this.description = description;
    infIf = new Vector();
    infThen = new Vector();
  }

  /** Create a new inference rule given its SGML tag representation. */
  public Def_Inference(SGML_Tag sgmltag) {

    description = sgmltag.getAttribute("DESCRIPTION");
    infIf = new Vector();
    infThen = new Vector();
  }

  public String getKey() {
  
    return description;
  }

  /** Returns the antecedents of the inference rule. */
  public Vector getInfIf() {

    return infIf;
  }

  /** Returns the consequents of the inference rule. */
  public Vector getInfThen() {

    return infThen;
  }

  /** Returns the description of the inference rule. */
  public String getDescription() {

    return description;
  }


  /*
  public void addElement(SHOE_Obj obj) throws DupeKeyException {
  }
  */

  /** Extract the INF-IF tags from the HTML tokenizer and add them
    to the rule. */
  public Vector readIfTags(HTML_Tokenizer intokens) {

    boolean endtag = false;
    int ttype;
    HTML_Token token;
    SGML_Tag sgmltag = null;
    String tagname;
    SHOE_Obj obj;
    SHOEException perr;
    Vector errStrings = null;

    errStrings = new Vector();
    do {
      ttype = intokens.nextToken();
      token = intokens.getLastToken();
      if (token instanceof SGML_Tag) {
	sgmltag = (SGML_Tag)token;
	tagname = sgmltag.getName();
	if (tagname.equals("CATEGORY")) {
	  Category cat = new Category(sgmltag);
	  cat.checkFields(sgmltag, errStrings);
	  infIf.addElement(cat);
	}
	else if (tagname.equals("RELATION")) {
	  Relation rel = new Relation(sgmltag);
	  rel.checkFields(sgmltag, errStrings);
	  rel.readSubTags(intokens);
	  infIf.addElement(rel);
	}
	else if (tagname.equals("COMPARISON")) {  
	  obj = new Comparison(sgmltag);
	  obj.readSubTags(intokens);
	  infIf.addElement(obj);	      
	}
	else if (tagname.equals("/INF-IF"))
	  endtag = true;
	else {                 // invalid tag
	  perr = new SHOEException(SHOEException.UNEXPECTED_TAG,
				   token.toString());
	  errStrings.addElement(perr.getMessage());
	}
      }
    } while (ttype != intokens.HT_EOF && endtag == false);
    if (ttype == intokens.HT_EOF) {
      perr = new SHOEException(SHOEException.NO_CLOSING_TAG, 
			       toString(), "</INF-IF>");
      errStrings.addElement(perr.getMessage());
    }
    return errStrings;
  }

  /** Extract the INF-THEN tags from the HTML tokenizer and add them
    to the rule. */
  public Vector readThenTags(HTML_Tokenizer intokens) {

    boolean endtag = false;
    int ttype;
    HTML_Token token;
    SGML_Tag sgmltag = null;
    String tagname;
    SHOE_Obj obj;
    SHOEException perr;
    Vector errStrings = null;

    errStrings = new Vector();
    do {
      ttype = intokens.nextToken();
      token = intokens.getLastToken();
      if (token instanceof SGML_Tag) {
	sgmltag = (SGML_Tag)token;
	tagname = sgmltag.getName();
	if (tagname.equals("CATEGORY")) {
	  obj = new Category(sgmltag);
	  infThen.addElement(obj);
	}
	else if (tagname.equals("RELATION")) {
	  obj = new Relation(sgmltag);
	  obj.readSubTags(intokens);
	  infThen.addElement(obj);
	}
	else if (tagname.equals("/INF-THEN"))
	  endtag = true;
	else {                 // invalid tag
	  perr = new SHOEException(SHOEException.UNEXPECTED_TAG,
				   token.toString());
	  errStrings.addElement(perr.getMessage());
	}
      }
    } while (ttype != intokens.HT_EOF && endtag == false);
    if (ttype == intokens.HT_EOF) {
      perr = new SHOEException(SHOEException.NO_CLOSING_TAG, 
			       toString(), "</INF-THEN>");
      errStrings.addElement(perr.getMessage());
    }
    return errStrings;
  }

  public Vector readSubTags(HTML_Tokenizer intokens) {
    
    boolean endtag = false;
    int ttype;
    HTML_Token token;
    SGML_Tag sgmltag = null;
    String tagname;
    SHOE_Obj obj;
    SHOEException perr;
    Vector errStrings;
    Vector subErrs;

    errStrings = new Vector();
    do {
      ttype = intokens.nextToken();
      token = intokens.getLastToken();
      if (token instanceof SGML_Tag) {
        sgmltag = (SGML_Tag)token;
        tagname = sgmltag.getName();
	if (tagname.equals("INF-IF")) {
	  subErrs = readIfTags(intokens);
	  if (subErrs.size() > 0) {
	    Enumeration enum = subErrs.elements();	  
	    errStrings.addElement("In IF part");
	    while (enum.hasMoreElements()) {
	      String errStr = "  " + (String)enum.nextElement();
	      errStrings.addElement(errStr);
	    }
	  }
	} 
	else if (tagname.equals("INF-THEN")) {
	  subErrs = readThenTags(intokens);
	  if (subErrs.size() > 0) {
	    Enumeration enum = subErrs.elements();	  
	    errStrings.addElement("In THEN part");
	    while (enum.hasMoreElements()) {
	      String errStr = "  " + (String)enum.nextElement();
	      errStrings.addElement(errStr);
	    }
	  }
	}
        else if (tagname.equals("/DEF-INFERENCE")) {
          endtag = true;
        }
	else {                 // invalid tag
	  perr = new SHOEException(SHOEException.UNEXPECTED_TAG,
				   token.toString());
	  errStrings.addElement(perr.getMessage());
	}
      }
    } while (ttype != intokens.HT_EOF && endtag == false);
    if (ttype == intokens.HT_EOF) {
      perr = new SHOEException(SHOEException.NO_CLOSING_TAG, 
			       toString(), "</DEF-INFERENCE>");
      errStrings.addElement(perr.getMessage());
    }
    return errStrings;
  }


  /** Verifies and internalizes the inference rule by validating
    each of its clauses. */
  public Vector validate(SHOE_Obj parent, OntManager ontman) {

    SHOE_Obj obj;
    Enumeration ifEnum;
    Enumeration thenEnum;
    Vector errStrings = null;
    Vector subErrs;

    // System.out.println("Validating inference rule " + toString());

    errStrings = new Vector();

    ifEnum = infIf.elements();
    while (ifEnum.hasMoreElements()) {
      obj = (SHOE_Obj)ifEnum.nextElement();
      if (obj instanceof Relation) {
	// below we pass the parent ontology for reference by the relation
	subErrs = ((Relation)obj).validate(parent, ontman);
	SHOEException.addSubErrors(errStrings, "In Relation " + obj.getKey(),
				   subErrs);
      }
      else if (obj instanceof Category) {
	// below we pass the parent ontology for reference by the category
	subErrs = ((Category)obj).validate(parent, ontman);
	SHOEException.addSubErrors(errStrings, "In Category " + obj.getKey(),
				   subErrs);
      }
      else if (obj instanceof Comparison) {
	// Note: do we need to validate Comparison clauses?
	;
      }
    }

    thenEnum = infThen.elements();
    while (thenEnum.hasMoreElements()) {
      obj = (SHOE_Obj)thenEnum.nextElement();
      if (obj instanceof Relation) {
	// below we pass the parent ontology for reference by the relation
	subErrs = ((Relation)obj).validate(parent, ontman);
	SHOEException.addSubErrors(errStrings, "In Relation " + obj.getKey(),
				   subErrs);
      }
      else if (obj instanceof Category) {
	// below we pass the parent ontology for reference by the category
	subErrs = ((Category)obj).validate(parent, ontman);
	SHOEException.addSubErrors(errStrings, "In Category " + obj.getKey(),
				   subErrs);
      }
    }
    
    return errStrings;
  }

  public void printSGML(PrintStream out) {
  
    Enumeration enum;
    SHOE_Obj obj;

    out.println(this);
    out.println("<INF-IF>");
    enum = infIf.elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      obj.printSGML(out);
    }
    out.println("</INF-IF>");

    out.println("<INF-THEN>");
    enum = infThen.elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      obj.printSGML(out);
    }
    out.println("</INF-THEN>");
    out.println("</DEF-INFERENCE>");
  }
 
  /** Returns the string representation of the tag. Any contents or end
    tags are not included.  */
  public String toString() {
  
    String str;
    
    str = "<DEF-INFERENCE ";
    if (description != null)
      str = str + " DESCRIPTION=\"" + description + "\"";
    str = str + ">";
    return str;
  }
}
