/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/*
  Change History:
  01/10/01, JDH: Changed constructor to take array instead of Vector
                 Added makeCopy()  and setName() methods
		 Changed short_para to shortPhrase
  01/02/01, JDH: Fixed getDefArgs() to sort the args by POS number
  11/01/00, JDH: Use accessor methods to get fields in Def_Arg
  10/26/00, JDH: Added accessor methods for name, description and short_para
                 Made name, description and short_para private
                 Added getDefArgTypeClass() method
  */

package SHOE;

import java.util.*;
import java.io.*;
import HTML.*;

/** This class implements facilities for defining a SHOE relation. */
public class Def_Relation extends SHOE_Obj {
	
  private String name;
  private String description;
  private String shortPhrase;
  private Vector argList;            // vector of Def_Arg

  /** Create a new relation definition given a name, description, short
    phrase and array of argument definitions. */
  public Def_Relation(String name, String description, String shortPhrase,
		      Def_Arg[] argArray)  {

    this.name = name;
    this.description = description;
    this.shortPhrase = shortPhrase;
    argList = new Vector(argArray.length);
    for (int i=0; i < argArray.length; i++)
      argList.addElement(argArray[i]);
  }
  
  /** Create a new relation definition given a name, description, short
    phrase and vector of argument definitions. This constructor is
    deprecated and future use should be avoided. */

  /*
  public Def_Relation(String name, String description, String shortPhrase,
		      Vector argList)  {

    this.name = name;
    this.description = description;
    this.shortPhrase = shortPhrase;
    this.argList = argList;
  }
  */

  /** Create a new relation definition from its SGML tag representation. */
  public Def_Relation(SGML_Tag sgmltag) {
 
     name = sgmltag.getAttribute("NAME");
     description = sgmltag.getAttribute("DESCRIPTION");
     shortPhrase = sgmltag.getAttribute("SHORT");
     argList = new Vector();
  }


  /** Returns a description of the relation. */
  public String getDescription() {

    return description;
  }

  public String getKey() {
  
    return name;
  }

  /** Returns the name of the relation. */
  public String getName() {

    return name;
  }

  /** Returns a short phrase that describes the relation. This is
    the SHORT field in the DEF-RELATION tag. */
  public String getPhrase() {

    return shortPhrase;
  }

  /** Returns the argument that is located at index idx in the list.
   Note that it is dangerous to use this method unless it is known
   that the arguments are sorted by position number. */
  public Def_Arg getDefArgAt(int idx) {

    return (Def_Arg)argList.elementAt(idx);
  }

  /** Returns the argument with POS field of pos. Note: if we sort the
    arguments, then we can have one-shot access. */
  public Def_Arg getDefArg(int pos) {

    int numarg;
    Def_Arg darg = null;
    int i = 0;
    boolean found = false;

    numarg = argList.size();
    while (i < numarg && found == false) {
      darg = (Def_Arg)argList.elementAt(i);
      if (darg.getPosition() == pos)
	found = true;
      i++;
    }
    if (found == true)
      return darg;
    else
      return null;
  }

  /** Returns an array of the Def-Relation's arguments, sorted by POS
    number. */
  public Def_Arg[] getDefArgs() {

    int i;
    int numarg;
    Def_Arg args[];
    Def_Arg darg;
    int pos;

    numarg = argList.size();
    args = new Def_Arg[numarg];
    for (i=0; i < numarg; i++) {
      darg = (Def_Arg)argList.elementAt(i);
      pos = darg.getPosition();
      if (pos < 1 || pos > numarg) {
	System.out.println("Error in Def_Relation.getDefArgs(): " +
			   "Argument POS doesn't correspond to arity");
	return null;
      }
      args[pos - 1] = darg;
    }
    return args;
  }


  /** Returns the required type class of the argument in the desired
    position. */
  public int getDefArgTypeClass(int pos) {

    Def_Arg      defArg;

    defArg = getDefArg(pos);
    if (defArg != null)
      return defArg.getTypeClass();
    else
      return Def_Arg.UNDEFINED;
  }

  /** Return the number of arguments defined for the relation. */
  public int getArity() {

    return argList.size();
  }

  /** If the relation has a SHORT name, it is returned, otherwise the
    name of the relation is returned */
  public String getLabel() {

    String label;

    if (shortPhrase != null)
      label = shortPhrase;
    else
      label = name;
    return label;
  }


  /** Returns a copy of the Def_Relation. This copy can be altered without
    changing the original. */
  public Def_Relation makeCopy() {

    Def_Relation drelCopy;
    Def_Arg[]    argsOrig;
    Def_Arg[]    argsCopy;
 
    argsOrig = getDefArgs();
    argsCopy = new Def_Arg[argsOrig.length];
    for (int i=0; i < argsCopy.length; i++) 
      argsCopy[i] = argsOrig[i].makeCopy();
    drelCopy = new Def_Relation(name, description, shortPhrase, argsCopy);
    return drelCopy;
  }


  public Vector readSubTags(HTML_Tokenizer intokens) {

    boolean endtag = false;
    int ttype;
    HTML_Token token;
    SGML_Tag sgmltag = null;
    String tagname;
    Def_Arg darg;
    SHOEException perr;
    Vector errStrings = null;

    errStrings = new Vector();    
    do {
      ttype = intokens.nextToken();
      token = intokens.getLastToken();
//      System.out.println("Token: " + token);
      if (token instanceof SGML_Tag) {
        sgmltag = (SGML_Tag)token;
//        System.out.println("SGML: " + sgmltag);
        tagname = sgmltag.getName();
        if (tagname.equals("/DEF-RELATION"))
          endtag = true;
        else if (tagname.equals("DEF-ARG")) {
	  darg = new Def_Arg(sgmltag);
          argList.addElement(darg);
	  if (darg.getPosition() == 0) {
	    perr = new SHOEException(SHOEException.MISSING_FIELD, 
				     sgmltag.toString(), "POS");
	    errStrings.addElement(perr.getMessage());
	  }
	  if (darg.getType() == null) {
	    perr = new SHOEException(SHOEException.MISSING_FIELD,
				     sgmltag.toString(), "TYPE");
	    errStrings.addElement(perr.getMessage());
	  }

        }
	else {                 // invalid tag
	  perr = new SHOEException(SHOEException.UNEXPECTED_TAG,
				   token.toString());
	  errStrings.addElement(perr.getMessage());
	}
      }
    } while (ttype != intokens.HT_EOF && endtag == false);
    if (ttype == intokens.HT_EOF) {
      perr = new SHOEException(SHOEException.NO_CLOSING_TAG, 
			       toString(), "</DEF-RELATION>");
      errStrings.addElement(perr.getMessage());
    }
    return errStrings;
  }


  /** Sets the name of the relation. */
  public void setName(String name) {

    this.name = name;
  }

  /** Validates each Def_Arg contained in the Def_Relation */
  public Vector validate(SHOE_Obj parent, OntManager ontman) {

    Vector errStrings;
    Enumeration enum;
    Def_Arg darg;
    Vector subErrs;

    errStrings = new Vector();

    enum = argList.elements();
    while (enum.hasMoreElements()) {
      darg = (Def_Arg)enum.nextElement();
      subErrs = darg.validate(parent, ontman);
      SHOEException.addSubErrors(errStrings, "In Argument " + 
				 darg.getPosition(), subErrs);
    }
    return errStrings;
  }

  public void printSGML(PrintStream out) {
  
    Enumeration enum;
    Def_Arg darg;
        
    out.println(this);
    enum = argList.elements();
    while (enum.hasMoreElements()) {          /* print arguments */
      darg = (Def_Arg)enum.nextElement();
      out.print("   ");
      darg.printSGML(out);
    }
    out.println("</DEF-RELATION>");
  }

  /** Returns the string representation of the tag. Any contents
   or end tags are not included. */
  public String toString() {
  
    String str;
    
    str = "<DEF-RELATION NAME=\"" + name + "\"";

    if (description != null)
      str = str + " DESCRIPTION=\"" + description + "\"";
    if (shortPhrase != null)
      str = str + " SHORT=\"" + shortPhrase + "\"";

    str = str + ">";
    return str;
  }
}

