/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/* Change History:
   01/29/01, JDH: getUseOnts() now returns an array of Use_Ontology
   11/03/00, JDH: Made all fields private.
             JDH: Changed constructor to take String array instead of Vector
*/

package SHOE;

import java.util.*;
import java.io.*;
import HTML.*;
import Tree.*;
import java.lang.System;

/** A class for representing a SHOE instance. Methods are provided to read
  the instance from a SHOE document, to write it back out, and to 
  manipulate an intermediate data structure. */
public class Instance extends SHOE_Obj {

  private String key;
  private Vector delegateto;
  private Hashtable use_onts;       /* contains Use_Ontology objects */
  private Vector declarations;   /* contains Category and Relation objects */
  private Vector subinstances;   /* contains Instance objects */
  private SHOE_Doc parentDoc = null;   /* document containing this instance */

  // public Instance()  {}

  /** Constructs an instance and assigns it the given key and set of
    instances that have been delegated to. */
  public Instance(String key, String[] delegates)  {
	
    this.key = key;

    delegateto = new Vector();
    for (int i=0; i < delegates.length; i++) 
      delegateto.addElement(delegates[i]);
    this.use_onts = new Hashtable();
    this.declarations = new Vector();
    this.subinstances = new Vector();
    contents = new Vector();
  }
	
  /** Constructs an instance and initializes it using the information
    provided in the specified SGML tag. */
  public Instance(SGML_Tag sgmltag) {
  
    StringTokenizer tknzr;
    String delstr;
    
    key = sgmltag.getAttribute("KEY");
    delstr = sgmltag.getAttribute("DELEGATE-TO");
    delegateto = new Vector();
    if (delstr != null) {
      tknzr = new StringTokenizer(delstr);
      while (tknzr.hasMoreTokens())
        delegateto.addElement(tknzr.nextToken());
    }
    use_onts = new Hashtable();
    declarations = new Vector();
    subinstances = new Vector();
    contents = new Vector();
  }

  /** Special constructor to identify the containing document of this
    instance. It is only needed so that the instance can add its
    nested instances as top-level instances in the document (part
    of a kludge to handle nested instances.
    */
  public Instance(SGML_Tag sgmltag, SHOE_Doc parentDoc) {
  
    StringTokenizer tknzr;
    String delstr;
    
    this.parentDoc = parentDoc;
    key = sgmltag.getAttribute("KEY");
    delstr = sgmltag.getAttribute("DELEGATE-TO");
    delegateto = new Vector();
    if (delstr != null) {
      tknzr = new StringTokenizer(delstr);
      while (tknzr.hasMoreTokens())
        delegateto.addElement(tknzr.nextToken());
    }
    use_onts = new Hashtable();
    declarations = new Vector();
    subinstances = new Vector();
    contents = new Vector();
  }

  /** Returns the unique identifier for the instance. */
  public String getKey() {
  
    return key;
  }

  /** Returns an array of the keys of instances to which this instance
    has delegated authority. */
  public String[] getDelegates() {

    String[] delArray;

    delArray = new String[delegateto.size()];
    delegateto.copyInto(delArray);
    return delArray;
  }


  /** Returns a string that is a white-space delimited concatenation
    of all of the delegate to elements. */
  public String getDelegateStr() {

    Enumeration enum;
    String str = "";

    enum = delegateto.elements();
    while (enum.hasMoreElements()) 
      str = str + enum.nextElement() + " ";
    return str;
  }

  /** Returns an array of the Use-Ontologies included in this
    instance. */
  public Use_Ontology[] getUseOnts() {

    Use_Ontology[] useOntArray;
    Enumeration    enum;
    Use_Ontology   uont;
    int            i = 0;

    useOntArray = new Use_Ontology[use_onts.size()];
    enum = use_onts.elements();
    while (enum.hasMoreElements()) {
      uont = (Use_Ontology)enum.nextElement();
      useOntArray[i] = uont;
      i++;
    }
    return useOntArray;
  }

  /** Returns an enumeration of the category and relation claims made
    by this instance. */
  public Enumeration getDeclarations() {

    return declarations.elements();
  }

  /** Returns the Use-Ontology object in which the specified prefix
    was assigned to a referenced ontology. */
  public Use_Ontology getUseOnt(String prefix) {

    return (Use_Ontology)use_onts.get(prefix);
  }

  /** Returns the set of use ontologies that occur within the instance.
    Each use ontology is the root of a tree where children in the
    tree represent use ontologies that occur in the ontology referenced
    by the parent use ontology. Each node consists of a Use-Ontology
    object and a label that is the prefix chain to reach that object.
  */
  public Vector getUseOntTreeList(OntManager ontman) {

    Vector           treeList;
    Enumeration      uontEnum;
    Use_Ontology     uont;
    LabeledTreeNode  uontNode;
    DefaultTreeModel nextTree;

    treeList = new Vector(use_onts.size());
    uontEnum = use_onts.elements();
    while (uontEnum.hasMoreElements()) {
      uont = (Use_Ontology)uontEnum.nextElement();
      uontNode = uont.getUseOntNode(ontman);
      nextTree = new DefaultTreeModel(uontNode);
      treeList.addElement(nextTree);
    }
    return treeList;
  }

  /** Set the key and delegation vector to the specified values. */
  public void update(String key, Vector delegateto)  {
	
    this.key = key;
    this.delegateto = delegateto;
  }

  /** Adds a new use ontology to the instance. Throws an exception
    if the prefix used has already been assigned to another ontology. */
  public void addUseOnt(Use_Ontology uont) throws DupeKeyException {
    
    addUseOnt(uont, true);
  }

  /** Adds a new use ontology to the instance and if the boolean
    is true, also updates the content vector of the instance.
    Throws an exception if the prefix used has already been assigned
    to another ontology. */
  public void addUseOnt(Use_Ontology uont, boolean addToContent) 
       throws DupeKeyException
  {
    String key;

    if (addToContent == true) 
      addSHOEtoContents(uont, SHOE_Obj.newline);   
    
    key = uont.getKey();
    if (key == null) {
      System.out.println("Use-Ontology " + uont + " has no prefix.");
      return;
    }
    if (use_onts.containsKey(key) == false)
      use_onts.put(key, uont);
    else
      throw new DupeKeyException();
  }

  /** Adds a new category or relation claim to the instance. */
  private void addDeclaration(SHOE_Obj dec) {

    addDeclaration(dec, true);
  }	

  /** Adds a new category or relation claim to the instance and if
    the boolean parameter is true, also updates the content vector
    of the instance. */
  private void addDeclaration(SHOE_Obj dec, boolean addToContent) {

    if (addToContent == true)
      addSHOEtoContents(dec, SHOE_Obj.newline);   
    declarations.addElement(dec);
  }

  /* The public versions of addDeclaration() require the additional
     OntManager parameter so that the definition of the element
     can be looked up in ontology and the declaration can be
     validated. */

  /** Adds a new category or relation claim to the instance and
    validates it. */
  public void addDeclaration(SHOE_Obj dec, OntManager ontman) {

    addDeclaration(dec, ontman, true);
  }

  /** Adds a new category or relation claim to the instance, validates
    it, and if the boolean parameter is true, also updates the content
    vector of the instance. */
  public void addDeclaration(SHOE_Obj dec, OntManager ontman, 
			     boolean addToContent)
  {
    if (addToContent == true) {
      addSHOEtoContents(dec, SHOE_Obj.newline);   
    }
    /* validate an object before adding it */
    if (dec instanceof Relation)
      ((Relation)dec).validate(this, ontman);
    else if (dec instanceof Category)
      ((Category)dec).validate(this, ontman);

    declarations.addElement(dec);
  }
  
  /** Add a new instance that is contained within the current instance.
    Note: Nested instances may go away in future versions of SHOE.
    This code does not allow subinstances to be accessed; they are
    actually treated as new top-level instances in the document. */
  public void addSubInst(Instance inst) throws DupeKeyException {

    subinstances.addElement(inst);
  }

  /** Removes the specified Use-Ontology from the instance. */
  public void deleteUseOnt(Use_Ontology uont) {

    use_onts.remove(uont.getKey());
    contents.removeElement(uont);
  }

  /** Removes the specified category or relation claim from the instance. */
  public void deleteDeclaration(Declaration dec) {

    int i;

    for (i=0; i < declarations.size(); i++)
      if (declarations.elementAt(i) == dec)
	declarations.removeElementAt(i);
    contents.removeElement(dec);

  }

  /** Reads and processes the tokens from an HTML_Tokenizer until 
    the instance closing tag is encountered.  Parse errors are
    permanently stored with the ontology in parseErrs. */
  public Vector readSubTags(HTML_Tokenizer intokens) {

    boolean endtag = false;
    int ttype;
    HTML_Token token;
    SGML_Tag sgmltag = null;
    String tagname;
    SHOE_Obj obj;
    SHOEException perr;
    Vector errStrings = null;
    Vector subErrs;

    errStrings = new Vector();
    do {
      ttype = intokens.nextToken();
      token = intokens.getLastToken();
      if (token instanceof SGML_Tag) {
        sgmltag = (SGML_Tag)token;
        tagname = sgmltag.getName();
        if (tagname.equals("/INSTANCE")) {
          endtag = true;	      
	  /* the end tag is not considered part of the contents */
	}
        else if (tagname.equals("USE-ONTOLOGY")) {
          obj = new Use_Ontology(sgmltag);
          try {
            addUseOnt((Use_Ontology)obj, false);
          } catch (DupeKeyException x) {
	    perr = new SHOEException(SHOEException.DUPLICATE_PREFIX,
				     obj.getKey());
	    errStrings.addElement(perr.getMessage());
	  }
	  contents.addElement(obj);
        }
        else if (tagname.equals("CATEGORY")) {
          Category cat = new Category(sgmltag);
	  cat.checkFields(sgmltag, errStrings);
          addDeclaration(cat, false);
	  contents.addElement(cat);
        }
        else if (tagname.equals("RELATION")) {
          Relation rel = new Relation(sgmltag);
	  rel.checkFields(sgmltag, errStrings);
          subErrs = rel.readSubTags(intokens);
	  SHOEException.addSubErrors(errStrings, "In Relation " + rel.getKey(),
				     subErrs);
          addDeclaration(rel, false);
	  contents.addElement(rel);
        }
        else if (tagname.equals("INSTANCE")) {
          obj = new Instance(sgmltag, parentDoc);
	  subErrs = obj.readSubTags(intokens);
	  SHOEException.addSubErrors(errStrings, "In Instance " + obj.getKey(),
				     subErrs);
	  // instead of adding instance as a child, we will create an
	  // equivalent top-level instance in the document
	  /*
          addSubInst(obj);
	  contents.addElement(obj);
	  */
	  if (parentDoc != null)
	    parentDoc.addInstance((Instance)obj);
	  /* BUG: the flattened child instance should inherit the
	     USE-ONTOLOGY statements of its parent. */
        }
	else {                 // invalid tag
	  perr = new SHOEException(SHOEException.UNEXPECTED_TAG,
				   token.toString());
	  errStrings.addElement(perr.getMessage());
	  contents.addElement(token);
	}
      }
      else                             // not an SGML tag
	contents.addElement(token);
    } while (ttype != intokens.HT_EOF && endtag == false);
    if (ttype == intokens.HT_EOF) {
      perr = new SHOEException(SHOEException.NO_CLOSING_TAG, 
			       toString(), "</INSTANCE>");
      errStrings.addElement(perr.getMessage());
    }
    return errStrings;
  }

  /** Checks the instance for reference errors. These are permanently stored
     in the parseErrs vector. */
  /* BUG: This does not automatically update the parseErrs for the
     document in which the Instance is contained. */
  public Vector validate(SHOE_Obj parent, OntManager ontman) {

    SHOE_Obj obj;
    Enumeration enum;
    Vector errStrings = null;
    Vector subErrs;
    // Enumeration errEnum;
    SHOEException perr;

    errStrings = new Vector();
    // System.out.println ("Validating instance");
    enum = declarations.elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Relation) {
	subErrs = ((Relation)obj).validate(this, ontman);
	SHOEException.addSubErrors(errStrings, "In Relation " + obj.getKey(),
				   subErrs);
      }
      else if (obj instanceof Category) {
	subErrs = ((Category)obj).validate(this, ontman);
	SHOEException.addSubErrors(errStrings, "In Category " + obj.getKey(),
				   subErrs);
      }
    }

    /* Validate all ontologies used by this instance */
    Use_Ontology uont;
    Ontology ont;
    String header;

    enum = use_onts.elements();
    while (enum.hasMoreElements()) {
      uont = (Use_Ontology)enum.nextElement();
      ont = ontman.getOntology(uont);
      if (ont == null) {
	perr = new SHOEException(SHOEException.NO_ONTOLOGY_FILE,
				 uont.toString());
	errStrings.addElement(perr.getMessage());
      }
      else {
	if (ont.hasParseErrors()) {
	  header = "Found errors in ontology " + ont.getOntLabel() + 
	    " at " + ont.getURL();
	  SHOEException.addSubErrors(errStrings, header, ont.parseErrs);
	}
      }
    }

    return errStrings;
  }


  /** Writes the instance and its components to the given print
    stream in SGML format. */
  public void printSGML(PrintStream out) {
  
    Enumeration enum;
    SHOE_Obj obj;
    
    out.print(this);
    printContents(out);
    out.print("</INSTANCE>");
  }

  /** Returns a string consisting of the corresponding open tag for
     the instance. */
  public String toString() {
  
    String str;
    
    str = "<INSTANCE KEY=\"" + key + "\"";
    if (delegateto.isEmpty() == false)
      str = str + " DELEGATE-TO=\"" + delegateto + "\"";
    str = str + ">";
    return str;
  }

}
