/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/*
  Change History:
    01/10/01, JDH: Use the changed Def_Relation constructor
    01/04/01, JDH: Setting a proxy URL doesn't delete previous OntRecord
    11/01/00, JDH: Use accessor methods for fields in Def_Rename
    10/26/00, JDH: Use accessor methods for fields in Ontology, Use-Ontology
                   and Unique_Elt
*/

package SHOE;

import java.util.*;
import java.net.*;
import java.io.*;
import java.lang.System;
import HTML.AltProxyUrl;

/* eventually want to add LRU policy to remove ontologies from cache */

/** A layer that caches ontologies so that they do not always have to
  be re-loaded from the web. The base-ontology is automatically loaded. */
public class OntManager extends Object {

  /** show trace when loading onts? */
  protected boolean ontTrace = false;    

  static final String BASE_ONT_ID = "base-ontology";
  static final String BASE_ONT_VER = "1.0";
  static String BASE_ONT_URL = "http://www.cs.umd.edu/projects/plus/SHOE/onts/base1.0.html";
  static String BASE_ONT_DESC = "The base SHOE ontology";

  static final String BASE_TYPES[][] = {
    {"STRING", "An HTML 2.0 String Literal constant", "A string"},
    {"NUMBER", "A floating-point numerical constant", "A number"},
    {"DATE", "An RFC 1123 Timestamp, as shown in section 3.3.1 of the HTTP/1.0 spec", "A date/time stamp"},
    {"TRUTH", "A boolean YES or NO value", "YES or NO"}};

  static final String BASE_CATS[][] = {
    {"Entity", null, "Entity", "The preferred ancestor category of all SHOE categories.  Do not subcategorize from this category--use SHOEEntity instead."},
    {"SHOEEntity", "Entity", "SHOE Entity", "The preferred ancestor category of all SHOE categories defined outside this ontology."}};

  static final String BASE_RELS[][] = {
    {"description", "is described as", "This is a human-readable description of a particular instance."},
    {"name", "is called", "A descriptive name for the instance.  This should be a (preferably singular or mass) noun or correctly cased proper noun."}};

  static final String BASE_ARGS[][][] = {
    // DEF_ARGS for description relation
    {{"1", "Entity", "Entity"}, {"2", "STRING", "Entity Description"}},
    // DEF_ARGS for name relation
    {{"1", "Entity", "Entity"}, {"2", "STRING", "Entity Name"}}};

  private Hashtable ontIndex;        // contains all loaded ontologies
  /*
    private Vector newestOnts;         // the ontologies most recently loaded
    */

  /** Constructs a new ontology manager with only the base ontology
    loaded. */
  public OntManager() {

    ontIndex = new Hashtable();
    // newestOnts = new Vector();
    initBaseOnt();                  // initialize with the base ontology
  }

  /** Stores the base ontology in the index. This way, there is no need to
    retrieve it from the web. */
  private void initBaseOnt() {

    Ontology base;
    Def_Type dtype;
    Def_Category dcat;
    Def_Relation drel;
    int i, j, k;
    int pos = 0;
    Vector isavec;
    Def_Arg[] argList;

    base = new Ontology(BASE_ONT_ID, BASE_ONT_VER, new Vector(),
			BASE_ONT_DESC, new Vector());

    /* add DEF_TYPEs */
    for (i=0; i < BASE_TYPES.length; i++) {
      dtype = new Def_Type(BASE_TYPES[i][0], BASE_TYPES[i][1],
			   BASE_TYPES[i][2]);
      try {
	base.addElement(dtype);
      } catch (Exception ex) {
	System.err.println("Couldn't add an element to the base ontology");
      }
    }
    
    /* add DEF_CATEGORYs */
    for (i=0; i < BASE_CATS.length; i++) {
      isavec = new Vector();
      if (BASE_CATS[i][1] != null)
	isavec.addElement(BASE_CATS[i][1]); 
      dcat = new Def_Category(BASE_CATS[i][0], isavec, BASE_CATS[i][3],
			   BASE_CATS[i][2]);
      try {
	base.addElement(dcat);
      } catch (Exception ex) {
	System.err.println("Couldn't add an element to the base ontology");
      }
    }
    
    /* add DEF_RELATIONS */
    for (i=0; i < BASE_RELS.length; i++) {
      argList = new Def_Arg[BASE_ARGS[i].length];
      for (j=0; j < BASE_ARGS[i].length; j++) {
	try {
	  pos = Integer.parseInt(BASE_ARGS[i][j][0]);
	} catch (NumberFormatException ex) {
	  System.err.println("Error! Couldn't parse position number");
	}
	argList[j] = new Def_Arg(pos, BASE_ARGS[i][j][1], BASE_ARGS[i][j][2]);
      }
      drel = new Def_Relation(BASE_RELS[i][0], BASE_RELS[i][2],
			      BASE_RELS[i][1], argList);
      try {
	base.addElement(drel);
      } catch (Exception ex) {
	System.err.println("Couldn't add an element to the base ontology");
      }
    }

    base.setURL(BASE_ONT_URL);
    base.validate(null, this);      // records any errors found
    storeOntology(base);
  }


  /** Sets up a proxy URL to use when retrieving a particular ontology. This
     is particularly useful if the OntManager is being used when there is
     no internet connection or a mirror site has been set up for the
     ontology. */
  public void setOntProxy(String id, String version, String proxy) {

    String key;
    OntRecord orec;

    key = Ontology.makeKey(id, version);
    // if the record already exists, then just update it with proxy info
    orec = (OntRecord)ontIndex.get(key);
    if (orec != null)
      orec.setProxyUrl(proxy);
    else {
      orec = new OntRecord(id, version, proxy);
      ontIndex.put(key, orec);
    }
  }


 /** Loads a set of ontology proxies from an input stream. Each line must
   be a white-space delimited list of ontology id, version and a proxy URL for
   retrieving it. If there are any error messages, they are concatenated
   and returned as a String. */
  public String loadOntProxyInputStream(InputStream inStream) {

    String errMsg = "";
    DataInputStream dataStream;
    boolean eof = false;
    String line;
    StringTokenizer tokened;
    String id;
    String version;
    String urlStr;
    URL url;
    int lineNo = 0;

    
    dataStream = new DataInputStream(inStream);
    try {
      while (!eof) {
	line = dataStream.readLine();
	if (line == null || line.equals(""))
	  eof = true;
	else {
	  lineNo++;
	  tokened = new StringTokenizer(line);
	  if (tokened.countTokens() != 3) {
	    errMsg = errMsg + "Line " + lineNo + 
                     ": Does not have three values " +
	             "(ontology, version, and URL)" + SHOE_Obj.newline;
	  }
	  else {
	    id = tokened.nextToken();
	    version = tokened.nextToken();
	    urlStr = tokened.nextToken();
	    try {
	      url = new URL(urlStr);
	    } catch (MalformedURLException ex) {
	      errMsg = errMsg + "Line " + lineNo + ": " +
	               urlStr + " is not a valid URL." + SHOE_Obj.newline;
	    }
	    setOntProxy(id, version, urlStr);
	  }
	}
      }
    } catch (IOException ex) {
      errMsg = errMsg + "Error reading at  line " + lineNo + SHOE_Obj.newline;
    }
    if (errMsg.equals(""))
      errMsg = null;
    return errMsg;
  }


  /** Retrieve an ontology based only on its label string (this has
    a format of "[id], v.[version]").This can only be used if the 
    system has previously loaded the ontology or it knows where to
    locate it. */
  public Ontology getOntology(String label) {

    String id;
    String version;
    Ontology ont;
    int split;

    split = label.indexOf(", v.");
    id = label.substring(0, split);
    version = label.substring(split + 4);
    ont = getOntology(id, version);
    return ont;
  }

  /** Retrieve an ontology based only on its id and version number. This
    can only be used if the system has previously loaded the ontology or it
    knows where to locate it. */
  public Ontology getOntology(String id, String version) {

    String key;
    OntRecord orec;
    Ontology ont = null;

    key = Ontology.makeKey(id, version);
    orec = (OntRecord)ontIndex.get(key);
    if (orec != null) {                 // ontology is in index
      if (orec.isLoaded()) {
	ont = orec.getOntology();
      }
      else {                   // see if there is a known location for it
	if (orec.getProxyUrl() != null) {
	  ont = loadOntology(id, version, null, orec);
	  if (ont != null)
	    storeOntology(ont);
	}
      }
    }
    if (ont == null)
      System.err.println("Can't find ontology " + id + ", v." + version);
    return ont;
  }

  /** Retrieve an ontology based on its label and a suggested location. */
  public Ontology getOntology(String id, String version, String url) {

    String key;
    OntRecord orec;
    Ontology ont = null;

    key = Ontology.makeKey(id, version);
    orec = (OntRecord)ontIndex.get(key);
    if (orec != null) {                           // ontology is known
      if (orec.isLoaded()) {
	ont = orec.getOntology();
      }
    }

    // if ontology was not pre-loaded, try to load it
    if (ont == null) {
      ont = loadOntology(id, version, url, orec);
      if (ont != null)
	storeOntology(ont);
      else
	System.err.println("Can't find ontology " + id + ", v." + version);
    }

    return ont;
  }

  /** Returns the ontology that is described in the specified Use-Ontology
     object. */
  public Ontology getOntology(Use_Ontology uont) {

    Ontology ont;

    ont = getOntology(uont.getId(), uont.getVersion(), uont.getUrl());
    return ont;
  }

  /** Returns an enumeration of all ontologies that have been read */
  /*
  public Enumeration getOntologies() {

    return ontIndex.elements();
  }
  */

  /** Returns an enumeration of strings, where each is a label of an
    ontology known to the system. Currently, a known ontology is one
    that has been parsed or has a proxy assigned to it. */
  public Enumeration getKnownOntologies() {

    Vector ontlist;
    OntRecord orec;
    Enumeration enum;
    String label;

    ontlist = new Vector();

    /* since both loaded and proxy onts have index entries, we merely
       need to scan the ontIndex. Since the loaded and proxy properties
       are indicated in the same record, there is no need to check for
       duplicates. */
    enum = ontIndex.elements();
    while (enum.hasMoreElements()) {
      orec = (OntRecord)enum.nextElement();
      label = orec.getOntLabel();
      ontlist.addElement(label);
    }
    return ontlist.elements();
  }

  /** Given an ontology id and version, and the name of an element in the
    ontology, returns the corresponding SHOE element. (A element is a
    Def-Category, Def-Relation, Def-Constant, Def-Rename, or Def-Type) */
  public SHOE_Obj getElement(String ontId, String ontVer, String name) {

    Ontology ont;
    SHOE_Obj obj = null;

    ont = getOntology(ontId, ontVer);
    if (ont != null) {
      obj = ont.getElement(name);
    }
    return obj;
  }

  /** Given a SHOE element id, returns the corresponding SHOE element.
    (A element is a Def-Category, Def-Relation, Def-Constant, Def-Rename,
    or Def-Type) */
  public SHOE_Obj getElement(Unique_Elt elmt_id) {

    SHOE_Obj obj = null;

    if (elmt_id != null) {
      obj = getElement(elmt_id.getOntId(), elmt_id.getOntVer(),
		       elmt_id.getName());
    }
    return obj;
  }


  /** Returns all ontologies that have been retrieved since a given
    date. */
  /*
  public Vector getNewestOnts(Date startDate) {

    Vector result;
    Enumeration enum;
    OntRecord orec;
    Ontology ont;

    result = new Vector();
    enum = ontIndex.elements();
    while (enum.hasMoreElements()) {
      orec = (OntRecord)enum.nextElement();
      if (orec.isLoaded() && startDate.before(orec.getRefDate())) {
	ont = orec.ont;           
	// we don't use OntRecord.getOntology() because that will update
	// the refDate.
	result.addElement(ont);
      }
    }
    return result;
  }
  */

  /** Given an ontology id and version number, returns the default URL
    for it. The default URL is the location where the ontology was
    opened from, or a location specified by the user. */
  public String getDefaultUrl(String id, String version) {

    String key;
    OntRecord orec;

    key = Ontology.makeKey(id, version);
    orec = (OntRecord)ontIndex.get(key);
    if (orec != null) 
      return orec.getDefaultUrl();
    else
      return null;
  }

  /** Given an ontology id and version number, returns the default prefix
    for it. The default prefix is the prefix that was specified by the
    the user at some earlier point in time. */
  public String getDefaultPrefix(String id, String version) {

    String key;
    OntRecord orec;

    key = Ontology.makeKey(id, version);
    orec = (OntRecord)ontIndex.get(key);
    if (orec != null) 
      return orec.getDefaultPrefix();
    else
      return null;
  }

  /** Sets the default url for the given ontology id and version number
    to the value specified by url. */
  public void setDefaultUrl(String id, String version, String url) {

    String key;
    OntRecord orec;

    key = Ontology.makeKey(id, version);
    orec = (OntRecord)ontIndex.get(key);
    if (orec != null) 
      orec.setDefaultUrl(url);
  }

  /** Sets the default prefix for the given ontology id and version number
    to the value specified by prefix. */
  public void setDefaultPrefix(String id, String version, String prefix) {

    String key;
    OntRecord orec;

    key = Ontology.makeKey(id, version);
    orec = (OntRecord)ontIndex.get(key);
    if (orec != null) 
      orec.setDefaultPrefix(prefix);
  }

  /** Given an ontology, create a new record for it and stores it
    in the index. If there is already a record, but the ontology
    has not been loaded, then updates the record to reflect the
    loading of the ontology. */
  public void storeOntology(Ontology ont) {

    OntRecord orec;
    String key;

    key = ont.getKey();
    orec = (OntRecord)ontIndex.get(key);
    if (orec == null) {                // no existing record 
      orec = new OntRecord(ont);
      ontIndex.put(key, orec);
      // newestOnts.addElement(ont);       // add ont to end of the newest list
    }
    else {
      if (orec.isLoaded() == false) {     // ont is known, but not loaded
	orec.setOntology(ont);
	// newestOnts.addElement(ont);       // add ont to end of the newest list
      }
    }
  }

  
  /** Loads and returns the requested ontology. If a proxy has been set
    for the ontology, then it is read from this proxy, otherwise it
    is read from the URL given (either can be a file URL). The OntRecord
    argument may be null, as it is only used to determine if there is
    a proxy URL. The returned ontology is guaranteed to have been
    validated.
    */
  private Ontology loadOntology(String id, String version, String pubUrlStr,
				OntRecord orec) {

    SHOE_Doc ontDoc;
    String urlString;            // the url used to actually load the document
    AltProxyUrl url;
    InputStream inStream;
    Enumeration enum;
    Object obj;
    Ontology ont = null;
    boolean found = false;
    String key;
    String proxy;
    
    if (ontTrace == true)
      System.out.println("Loading ontology " + id + ", v. " + version);

    // does a proxy exist?
    key = Ontology.makeKey(id, version);
    urlString = pubUrlStr;
    if (orec != null) {
      proxy = orec.getProxyUrl();
      if (proxy != null) {
	urlString = proxy;
	/*
	  System.out.println("Using " + urlString + 
	              " as location for ontology " + id + ", v. " + version);
      */
      }
    }
    ontDoc = new SHOE_Doc();
    try {
      /* we use AltProxyUrl so when this is called from within an
	 applet, it will use the proxy server to fetch web pages from
	 other servers, and thus avoid Java applet security restrictions. */
      url = new AltProxyUrl(urlString);
      inStream = url.openStream();
      ontDoc.parse(inStream);
    } catch (MalformedURLException ex) {
      /*
	/////// Modified by Qin ////////////////////
      try {
        File ontfile= new File(urlString);
	  inStream = new FileInputStream(ontfile);
	  ontDoc.parse(inStream);
	  inStream.close();
      } catch (FileNotFoundException e) {
 	  System.out.println("Error opening URL or File");
      } catch (IOException e) {
 	  System.out.println("Error reading File");
	}
	////////////////////////////////////////////
	*/
      System.err.println("Error opening URL " + urlString);
    } catch (IOException ex) {
      System.err.println("Error reading ontology " + id + ", v." + version +
			 "from URL " + urlString);
      return null;
    }
    enum = ontDoc.getPSHOE();
    while (enum.hasMoreElements() && found == false){
      obj = enum.nextElement();
      if (obj instanceof Ontology) {
	ont = (Ontology)obj;
	/* make sure ont is the one requested */
	if (ont.getKey().equals(Ontology.makeKey(id, version))) {
	  ont.setURL(pubUrlStr);
	  ont.validate(null, this);      // records any errors found
	  found = true;
	}
      }
    }
    if (found == true)
      return ont;
    else {
      System.err.println("Couldn't find ontology " + id + ", v." + version + 
			 " in file " + urlString);
      return null;
    }
  }

  /** Loads the ontology specified by a Use-Ontology structure. This
     method is currently unused.  */
  private Ontology loadOntology(Use_Ontology uont) {

    Ontology ont = null;

    ont = loadOntology(uont.getId(), uont.getVersion(), uont.getUrl(), null);
    return ont;
  }

  /** Determines if the category identified by the first element id is
    a subcategory of the one identified by the second. Returns true
    or false. This routine performs a breadth-first search with the
    child as the start state and the parent as the goal. */
  /* NOTE: This method can be inefficient when there are multiple parents
     for a category, and these parents share many common ancestors, because
     each of the common ancestors will be visited twice if the
     desired parent is not an ancestor. */
  public boolean isSubcategoryOf(Unique_Elt childId, Unique_Elt parId) {

    Vector queue;
    Unique_Elt ancestorId;          // unique id of an ancestor of child
    SHOE_Obj ancestorObj;           // SHOE object for ancestor   
    Def_Category ancestorCat;       // Def_Category for ancestor
    Enumeration parIds;
    Unique_Elt aParentId;

    // return false if either argument is null
    if (childId != null && parId != null) {

      queue = new Vector();
      queue.addElement(childId);

      while (queue.isEmpty() == false) {
	ancestorId = (Unique_Elt)queue.firstElement();
	queue.removeElementAt(0);
	if (ancestorId.equals(parId))
	  return true;
	else {
	  ancestorObj = getElement(ancestorId);
	  if (ancestorObj instanceof Def_Category) {
	    ancestorCat = (Def_Category)ancestorObj;
	    parIds = ancestorCat.getParentIds();
	    while (parIds.hasMoreElements()) {
	      aParentId = (Unique_Elt)parIds.nextElement();
	      queue.addElement(aParentId);
	    }
	  }
	}
      }
    }
    return false;        // didn't find desired parent in ancestors
  }

  /** Given a prefixed name and a reference object in which it appears,
    follows the prefixes to determine the originating ontology and
    version for the named object, and uses this to construct and return
    a unique identifier for it. */
  public Unique_Elt resolveName(String name, SHOE_Obj refObj) 
       throws SHOEException
  {
    Unique_Elt elmt_id = null;
    String rname;
    String prefix;
    SHOE_Obj obj;
    Ontology ont;
    Instance inst;
    Use_Ontology useOnt = null;

    if (name == null)
      return null;

    rname = name;
    prefix = getPrefix(rname);
    // System.out.println("Prefix is " + prefix);

    if (prefix == null) {

      /* if no prefix, object must appear in the reference ontology */
      if (refObj instanceof Ontology) {
	ont = (Ontology)refObj;
	obj = ont.getElement(name);

	if (obj == null)
	  throw new SHOEException(SHOEException.NO_ELEMENT_DEF,
				  name, ont.getOntLabel());

	// if the object is a rename, return the id of the renamed object
	if (obj instanceof Def_Rename) {
	  Def_Rename drnm = (Def_Rename)obj;
	  elmt_id = drnm.getFromId();

	  /* if the reference is in the same ont as the rename definition
	     then the rename's element may not have been found yet */
	  if (elmt_id == null)
	    elmt_id = resolveName(drnm.getFrom(), refObj);
	  return elmt_id;
	}
	else {
	  elmt_id = new Unique_Elt(ont.getId(), ont.getVersion(), rname);
	  return elmt_id;
	}
      }
      else {
	throw new SHOEException(SHOEException.MISSING_PREFIX, name);
      }
    }

    if (prefix.equals("")) {
      /* if empty prefix chain (nothing before the '.', then assume the 
	 object comes from the base ontology */
      ont = getOntology(BASE_ONT_ID, BASE_ONT_VER, BASE_ONT_URL);
      if (ont == null)
	throw new SHOEException(SHOEException.MISSING_ONTOLOGY,
				BASE_ONT_ID + ", v." + BASE_ONT_VER);
    }
    else {
      /* get the use ontology for the first prefix from the reference object */
      if (refObj instanceof Instance) {
	inst = (Instance)refObj;
	useOnt = inst.getUseOnt(prefix);
      }
      else if (refObj instanceof Ontology) {
	ont = (Ontology)refObj;
	useOnt = ont.getUseOnt(prefix);
      }
      else {
	System.err.println("System error: Wrong type of reference object " +
			   "passed to resolveName()");
      }
      
      /* if couldn't find initial prefix, check parent instances for
	 valid use ontologies (must add this when sub-instances work!!!!!) */
      /*    if (useOnt == null && refObj instanceof Instance) {
	    while (use_ont == null && inst.getPartOf() != null) {
	    inst = (Instance)inst.partOf;
	    useOnt = inst.getUseOnt(prefix);
	    }
	    if (useOnt == null)
	    return null;            // throw an error!
	    }
	    */

      //    System.out.println("Resolve: useont = " + useOnt);

      if (useOnt == null)
	throw new SHOEException(SHOEException.PREFIX_NOT_DEFINED, prefix);

      /* get the ontology refered to by the prefix */
      ont = getOntology(useOnt); 
      if (ont == null)
	throw new SHOEException(SHOEException.MISSING_ONTOLOGY,
				useOnt.getKey());
    }

    /* resolve any remaining prefixes */
    rname = getPrefixRemainder(rname);

    //    System.out.println("Rest of name: " + rname);

    prefix = getPrefix(rname);
    while (prefix != null) {
      /* find prefix in ontology's USE-ONTOLOGY list */
      useOnt = ont.getUseOnt(prefix);
      if (useOnt == null)
	throw new SHOEException(SHOEException.PREFIX_NOT_DEFINED, prefix);
      ont = getOntology(useOnt);
      if (ont == null)
	throw new SHOEException(SHOEException.MISSING_ONTOLOGY,
				useOnt.getKey());
      rname = getPrefixRemainder(rname);
      prefix = getPrefix(rname);
    }

    /* check for remaining name in ontology */
    obj = ont.getElement(rname);
    if (obj == null)
      throw new SHOEException(SHOEException.NO_ELEMENT_DEF,
			      rname, ont.getOntLabel());

    if (obj instanceof Def_Rename)         // return elmt_id of renamed obj
      elmt_id = ((Def_Rename)obj).getFromId();
    else
      elmt_id = new Unique_Elt(ont.getId(), ont.getVersion(), rname);
    return elmt_id;
  }

  /** Given a string that represents an element reference, returns the
    first prefix in the string. */
  String getPrefix(String str) {
    
   int pend;

   pend = str.indexOf('.');
   if (pend < 0)
     return null;
   else 
     return str.substring(0, pend);
 }

  /** Given a string that represents an element reference, returns the
    portion of the string the follows the first prefix. */
  String getPrefixRemainder(String str) {
    
   int pend;

   pend = str.indexOf('.');
   if (pend < 0)
     return str;
   else 
     return str.substring(pend + 1, str.length());
  }


  /** Returns a list of Def-Relations in the ont ontology in which argument
    position argPos can be legally filled by the category identified
    by elmt_id. This list includes both locally defined relations and
    "virtual relations" for those relations defined in other ontologies
    but been given a local alias via a rename.  These "virtual relations"
    are just like ordinary Def-Relations except the the name has been
    changed to the alias given by the Def_Rename.
  */
  public Vector getRelsByArgType(Ontology ont, Unique_Elt elmt_id, int argPos)
  {
    Vector relList;
    Hashtable ancestorIds;
    Enumeration eltEnum;
    SHOE_Obj obj, baseObj;

    relList = new Vector();

    ancestorIds = getAncestorHashtable(elmt_id);
    
/* TEST
      // debugging code
      Unique_Elt e;
      Enumeration aenum = ancestorIds.elements();
      System.out.println("Ancestors of " + elmt_id.toString());
      while (aenum.hasMoreElements()) {
        e = (Unique_Elt)aenum.nextElement();
        System.out.println(e.toString());
      }
*/

    eltEnum = ont.getElements();
    while (eltEnum.hasMoreElements())
    {
      obj = (SHOE_Obj)eltEnum.nextElement();
      baseObj = obj;
      /* in the case of renames, find the original object */
      if (obj instanceof Def_Rename) {
         elmt_id = ((Def_Rename)obj).getFromId();
         baseObj = getElement(elmt_id);
      }
      
      /* below we check the type of the base object to determine if
         it is a relation. This way renamed relations will appear in
	 the list. Before a renamed relation is added to the
	 list, it is copied, and the name is changed to the local
	 alias provided by the rename.
      */
      
      if (baseObj instanceof Def_Relation) {
         Def_Relation drel = (Def_Relation)baseObj;

	 if (argPos > 0 && argPos <= drel.getArity()) {
	   Def_Arg subjectArg = drel.getDefArgAt(argPos - 1);
	   Unique_Elt typeId = subjectArg.getTypeId();
/* TEST
           System.out.println("typeId=" + typeId);
 */
	   if (ancestorIds.containsKey(typeId)) {
	     // if this is a renamed object, make a copy and change the name
	     if (obj instanceof Def_Rename) {
	       drel = drel.makeCopy();
	       drel.setName(((Def_Rename)obj).getTo());
	     }
	     relList.addElement(drel);
	   }
	 }
      }
    }
    return relList;
  }

  /** Given a hashtable and a id of a Def_Category, adds each parent of
     the category to the hashtable. Duplicates are not added. This
     method supports getAncestorHashtable. */
  public void putParentsInHashTable(Hashtable ancestors, Unique_Elt catId)
  {
    Def_Category dcat;
    Enumeration enum;
    Unique_Elt pid;           // id for parent category

    dcat = (Def_Category)getElement(catId);
    enum = dcat.getParentIds();
    while (enum.hasMoreElements()) {
      pid = (Unique_Elt)enum.nextElement();
      if (ancestors.containsKey(pid) == false) {
         ancestors.put(pid, "");
         putParentsInHashTable(ancestors, pid);
      }
    }
  }

  /** Returns a hashtable consisting of the ids of all categories
    (including those in other ontologies) of which the input argument
    is a sub-category. The input should be a Def_Category, Def_Constant
    or a rename of one. If the input is a Def_Category, it is included in
    the results.
  */
  public Hashtable getAncestorHashtable(Unique_Elt elmt_id)
  {
    Hashtable ancestorIds = null;
    SHOE_Obj obj;
    Unique_Elt catId;
    Unique_Elt fromId;

    obj = getElement(elmt_id);
    if (obj instanceof Def_Rename) {
      fromId = ((Def_Rename)obj).getFromId();
      ancestorIds = getAncestorHashtable(fromId);
    }
    else if (obj instanceof Def_Constant) {
      catId = ((Def_Constant)obj).getCategoryId();
      ancestorIds = getAncestorHashtable(catId);
    }
    else if (obj instanceof Def_Category) {
      ancestorIds = new Hashtable();
      ancestorIds.put(elmt_id, "");
      putParentsInHashTable(ancestorIds, elmt_id);
    }
    else {
      System.err.println("ERROR: Called getAncestorHashtable on " + elmt_id);
    }
    return ancestorIds;
  }

}

/** The objects that are stored in ontIndex. As such, it allows additional
  information to be recorded for each ontology. */
class OntRecord {

  String   id;
  String   version;
  String   proxyUrl;
  Ontology ont;                // null if ontology not currently loaded
  Date     loadDate;           // date when the ontology was first loaded
  Date     refDate;            // date when the ontology was last referenced
  private String   defaultUrl;    // default URL for a Use-Ontology statement
  String   defaultPrefix;      // default prefix for a Use-Ontology statement

  /** Constructor for creating a new record from a parsed ontology. */
  public OntRecord(Ontology ont) {

    this.id = ont.getId();
    this.version = ont.getVersion();
    this.proxyUrl = null;
    this.ont = ont;
    this.loadDate = new Date();
    this.refDate = this.loadDate;     // retrieval date is same as load date
    this.defaultUrl = ont.getURL();
    this.defaultPrefix = null;
  }

  /** Constructor for creating a new record for a proxy URL. */
  public OntRecord(String id, String version, String proxyUrl) {

    this.id = id;
    this.version = version;
    this.ont = null;
    this.loadDate = null;
    this.refDate = null;
    this.proxyUrl = proxyUrl;
    this.defaultUrl = null;
    this.defaultPrefix = null;
  }

  public Ontology getOntology() {

    refDate = new Date();             // update the last date of retrieval
    return ont;
  }

  public String getOntLabel() {

    return id + ", v." + version;
  }

  /** Returns the date that the ontology was loaded. */
  public Date getLoadDate() {
    
    return loadDate;
  }

  /** Returns the date that the ontology was last retreived. */
  public Date getRefDate() {

    return refDate;
  }

  public String getProxyUrl() {

    return proxyUrl;
  }

  public String getDefaultUrl() {

    return defaultUrl;
  }

  public String getDefaultPrefix() {

    return defaultPrefix;
  }

  public boolean isLoaded() {

    if (ont == null)
      return false;
    else
      return true;
  }

  public void setDefaultPrefix(String defaultPrefix) {

    this.defaultPrefix = defaultPrefix;
  }

  public void setDefaultUrl(String defaultUrl) {

    this.defaultUrl = defaultUrl;
  }

  /** Update the record with a copy of the ontology. Also update the
    time that it was loaded and last referenced, as well as the URL
    from whence it came. */
  public void setOntology(Ontology ont) {
    
    this.ont = ont;
    loadDate = new Date();             // update the load date
    refDate = loadDate;                // update the last retrieval date
    setDefaultUrl(ont.getURL());       // update the default location
  }

  public void setProxyUrl(String proxyUrl) {

    this.proxyUrl = proxyUrl;
  }

}





