/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/*
  Change History:
    01/09/00, JDH: Changed getRelations() to return an array of Def_Relations
    12/29/00, JDH: getUseOnts() now returns an array of Use_Ontology
    11/03/00, JDH: Made all fields private
    10/26/00, JDH: Added accessor method for description
*/

package SHOE;

import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.*;
import java.lang.System;
import HTML.*;
import Tree.*;

/** A SHOE ontology. Methods are provided to read the ontology from a
  SHOE document, to write it back out, and to manipulate an intermediate
  data structure. */
public class Ontology extends SHOE_Obj {

  private String id;
  private String version;
  private Vector back_compatible;
  private String description;
  private Vector declarators;

  private Hashtable use_onts;
  private Hashtable elements;
  private Vector inferences;

  private Hashtable localNameCats;         // categories with local names
  private Vector topLevelCats;             // top level categories

  /* this is provided so that the Ontology Manager can store the location
     of the ontology. This URL is used to identify the ontology if it
     has errors. */
  private String url = null;

  protected Vector parseErrs;    // records errors found in the ontology page

  /** Creates a unique key for an ontology from its id and version
    number. Currently, this key is the concatenation of the version and
    the id. The version is used first to ensure it is used in
    the standard string hashing function.   */
  public static String makeKey(String id, String version) {

    return version + id;
  }

  /** Creates a new ontology with the given id and  version number. */
  public Ontology(String id, String version)  {

    this.id = id;
    this.version = version;
    this.back_compatible = new Vector();
    this.description = null;
    this.declarators = new Vector();
    
    use_onts = new Hashtable();
    elements = new Hashtable();
    inferences = new Vector();

    localNameCats = new Hashtable();
    topLevelCats = new Vector();
    parseErrs = new Vector();
  }


  /** Creates a new ontology with the given id, version number,
    set of versions it is backward-compatible with, description,
    and set of instances that make declarations for it. */
  public Ontology(String id, String version, Vector back_compatible,
		  String description, Vector declarators)  {

    this.id = id;
    this.version = version;
    this.back_compatible = back_compatible;
    this.description = description;
    this.declarators = declarators;
    
    use_onts = new Hashtable();
    elements = new Hashtable();
    inferences = new Vector();

    localNameCats = new Hashtable();
    topLevelCats = new Vector();
    parseErrs = new Vector();
  }

  /** Creates a new ontology and initializes it using the information
    provided in the given SGML tag. */
  public Ontology(SGML_Tag sgmltag) {

    StringTokenizer tknzr;
    String backstr;
    String declstr;
    
//    System.out.println ("Ont Tag as SGML: " + sgmltag);
    
    id = sgmltag.getAttribute("ID");
    version = sgmltag.getAttribute("VERSION");
    description = sgmltag.getAttribute("DESCRIPTION");
    back_compatible = new Vector();
    declarators = new Vector();
    use_onts = new Hashtable();
    elements = new Hashtable();
    inferences = new Vector();
    
    localNameCats = new Hashtable();
    topLevelCats = new Vector();
    parseErrs = new Vector();

    backstr = sgmltag.getAttribute("BACKWARD-COMPATIBLE-WITH");
    declstr = sgmltag.getAttribute("DECLARATORS");
    if (backstr != null) {
      tknzr = new StringTokenizer(backstr);
      while (tknzr.hasMoreTokens())
        back_compatible.addElement(tknzr.nextToken());
    }
    if (declstr != null) {
      tknzr = new StringTokenizer(declstr);
      while (tknzr.hasMoreTokens())
        declarators.addElement(tknzr.nextToken());
    }
  }

  /** Returns a unique identifier for the ontology. */
  public String getKey() {
  
    return Ontology.makeKey(id, version);
  }

  /** Returns the id of the ontology. */
  public String getId() {

    return id;
  }

  /** Returns the version number of the ontology. */
  public String getVersion() {

    return version;
  }

  /** Returns a label suitable for identifying the ontology to a
    human reader. */
  public String getOntLabel() {

    return id + ", v." + version;
  }

  /** Return the description of the ontology. */
  public String getDescription() {
    
    return description;
  }

  /** Return an array of version numbers with which this version is
    backward compatible. */
  public String[] getVersionsCompatibleWith() {

    String[] compatVers;

    compatVers = new String[back_compatible.size()];
    back_compatible.copyInto(compatVers);
    return compatVers;
  }

  /** Return the string that consists of a whitespace separated list of
    versions with which this ontology is backward compatible. */
  public String getBackCompatString() {

    String str = "";
    int cnt;
    int i;

    cnt = back_compatible.size();
    for (i=0; i < cnt; i++)
      str = str + back_compatible.elementAt(i) + " ";
    str.trim();
    return str;
  }

  /** Return the string that consists of a whitespace separated list of
    the instances that make special declarations for this ontology. */
  public String getDeclaratorsString() {

    String str = "";
    int cnt;
    int i;

    cnt = declarators.size();
    for (i=0; i < cnt; i++)
      str = str + declarators.elementAt(i) + " ";
    str.trim();
    return str;
  }

  /** Returns an array of the Use_Ontology objects which represent
    the ontologies included in this ontology. */
  public Use_Ontology[] getUseOnts() {

    Use_Ontology[] useOntArray;
    Enumeration    enum;
    Use_Ontology   uont;
    int            i = 0;

    useOntArray = new Use_Ontology[use_onts.size()];
    enum = use_onts.elements();
    while (enum.hasMoreElements()) {
      uont = (Use_Ontology)enum.nextElement();
      useOntArray[i] = uont;
      i++;
    }
    return useOntArray;
  }

  /** Returns an enumeration of the ontology elements in an arbitrary
    order. */
  public Enumeration getElements() {

    return elements.elements();
  }

  /** Returns an enumeration of the ontology elements in sorted order. */
  public Enumeration getSortedElements() {

    Vector sortElts;

    sortElts = SHOE_Obj.sortObjectEnum(elements.elements());
    return sortElts.elements();
  }

  /** Returns an array of all DEF-RELATIONs defined in the ontology.
    This does not return other Def_Relations which can be referenced by
    the ontology either through local aliases or prefixed names.*/
  public Def_Relation[] getRelations() {

    Vector relVec;
    Enumeration defEnum;
    Def_Relation[] relArray;
    SHOE_Obj elmt;

    relVec = new Vector();
    defEnum = elements.elements();
    while (defEnum.hasMoreElements()) {
      elmt = (SHOE_Obj)defEnum.nextElement();
      if (elmt instanceof Def_Relation)
	relVec.addElement(elmt);
    }
    relArray = new Def_Relation[relVec.size()];
    relVec.copyInto(relArray);
    return relArray;
  }


  /** Returns the inference rules stated for the ontology. */
  public Vector getInferenceRules() {

    return inferences;
  }

  /** Returns the set of use ontologies that occur within the ontology.
    Each use ontology is the root of a tree where children in the
    tree represent use ontologies that occur in the ontology referenced
    by the parent use ontology. Each node consists of a Use-Ontology
    object and a label that is the prefix chain to reach that object.
  */
  public Vector getUseOntTreeList(OntManager ontman) {

    Vector           treeList;
    Enumeration      uontEnum;
    Use_Ontology     uont;
    LabeledTreeNode  uontNode;
    DefaultTreeModel nextTree;

    treeList = new Vector(use_onts.size());
    uontEnum = use_onts.elements();
    while (uontEnum.hasMoreElements()) {
      uont = (Use_Ontology)uontEnum.nextElement();
      uontNode = uont.getUseOntNode(ontman);
      nextTree = new DefaultTreeModel(uontNode);
      treeList.addElement(nextTree);
    }
    return treeList;
  }

  /** Returns the set of category hierarchies defined in the ontology
    as an enumeration. Each hierarchy is a tree with nodes that consist
    of the local name as a label and the original Def_Category as a
    user object. */
  public Enumeration getLabeledCategoryTreeLists(OntManager ontman) {

    Vector           treeList;
    Enumeration      topEnum;
    SHOE_Obj         nextTop;
    DefaultTreeModel nextTree;

    treeList = new Vector(topLevelCats.size());
    topEnum = topLevelCats.elements();
    while (topEnum.hasMoreElements()) {
      nextTop = (SHOE_Obj)topEnum.nextElement();
      nextTree = getLabeledDescendantTree(nextTop, ontman);
      treeList.addElement(nextTree);
    }
    return treeList.elements();
  }

  /** Returns a tree of the the category hierarchy rooted at the chosen
    Def-Category, where each node has a label representing the local
    name of the category as well as the category object. The children
    are stored in sorted order by key. */
  public DefaultTreeModel getLabeledDescendantTree(SHOE_Obj obj,
						   OntManager ontman) {

    SHOE_Obj         baseObj;
    Unique_Elt       elmt_id;
    Def_Category     dcat;
    LabeledTreeNode  root;
    DefaultTreeModel tree = null;

    baseObj = obj;
    if (obj instanceof Def_Rename) {
      /* in the case of renames, determine which object it references */
      elmt_id = ((Def_Rename)obj).getFromId();
      baseObj = ontman.getElement(elmt_id);
    }

    if (baseObj instanceof Def_Category) {
      dcat = (Def_Category)baseObj;
      root = getLabeledDescendantNode(dcat, obj.getKey(), ontman);
      tree = new DefaultTreeModel(root);
    }
    return tree;
  }


  /** Given a Def_Category, coumputes the subtree of its descendants that
    are defined or aliased in the current ontology and returns a node
    representing the corresponding subtree. The children are stored in
    sorted order by the label. */
  private LabeledTreeNode getLabeledDescendantNode(Def_Category aCat,
						   String label,
						   OntManager ontman) {

    LabeledTreeNode  newNode;
    LabeledTreeNode  childNode;
    Enumeration      childIds;
    Unique_Elt       nextChildId;
    SHOE_Obj         localObj;
    Def_Category     childCat = null;
    String           childName;
    Unique_Elt       elmt_id;

    newNode = new LabeledTreeNode(aCat, label);
    childIds = aCat.getChildIds();
    while (childIds.hasMoreElements()) {
      nextChildId = (Unique_Elt)childIds.nextElement();

      /* if the child has a local name, add it as a child in the
	 descendant tree */
      localObj = (SHOE_Obj)localNameCats.get(nextChildId);
      if (localObj != null) {
	if (localObj instanceof Def_Category) {
	  childCat = (Def_Category)localObj;
	}
	else if (localObj instanceof Def_Rename) {
	  elmt_id = ((Def_Rename)localObj).getFromId();
	  childCat = (Def_Category)ontman.getElement(elmt_id);
	} 
	else {
	  System.out.println("Internal error in getLabeledDescendantNode()");
	  return null;
	}
	childName = localObj.getKey();
	childNode = getLabeledDescendantNode(childCat, childName, ontman);
	newNode.insertInSortedOrder(childNode);
      }
    }
    return newNode;
  }

  /** Returns the URL that has been assigned to this ontology. This
    method is used in error handling to help identify the file involved. */
  public String getURL() {

    return url;
  }

  /** Sets the URL where the ontology was retrieved from. */
  protected void setURL(String url) {

    this.url = url;
  }

  /** Given a prefix, returns the Use-Ontology statement which 
    identifies the ontology referenced by the prefix. */
  public Use_Ontology getUseOnt(String prefix) {

    return (Use_Ontology)use_onts.get(prefix);
  }

  /** Returns the ontology object (i.e., Def_Category, Def_Relation,
     Def_Const, Def_Type, or Def_Rename) that has the given name.
  */
  public SHOE_Obj getElement(String name) {

    return (SHOE_Obj)elements.get(name);
  }

  /** Adds a Use-Ontology to the ontology. If the prefix to be
    assigned to referenced ontology is already in used, then an
    exception is thrown. */
  public void addUseOnt(Use_Ontology uont) throws DupeKeyException {
              
    String key;
    
    key = uont.getKey();
    if (use_onts.containsKey(key) == false)
      use_onts.put(key, uont);
    else
      throw new DupeKeyException();
  }

  /** Adds the definition of an ontology element to the ontology.
    If the name of the element is already used by another object
    in the ontology, an exception is thrown. */
  public void addElement(SHOE_Obj obj) throws DupeKeyException {

    String key;
    
    key = obj.getKey();
    if (key == null)
      key = "missingkey";
    // System.out.println("Adding key " + key);
    if (elements.containsKey(key) == false)
      elements.put(key, obj);
    else
      throw new DupeKeyException();
  }

  /** Add an inference rule to the ontology. */
  public void addInference(Def_Inference inf) {

    inferences.addElement(inf);
  }

  /** Reads and processes the tokens from an HTML_Tokenizer until 
    a closing ontology tag is encountered. Parse errors are
    permanently stored with the ontology in parseErrs. */
  public Vector readSubTags(HTML_Tokenizer intokens) {
    
    boolean endtag = false;
    int ttype;
    HTML_Token token;
    SGML_Tag sgmltag = null;
    String tagname;
    SHOE_Obj obj;
    SHOEException perr;
    Vector subErrs;

    do {
      ttype = intokens.nextToken();
      token = intokens.getLastToken();
      if (token instanceof SGML_Tag) {
        sgmltag = (SGML_Tag)token;
        tagname = sgmltag.getName();
        if (tagname.equals("/ONTOLOGY"))
          endtag = true;
        else if (tagname.equals("USE-ONTOLOGY")) {
          obj = new Use_Ontology(sgmltag);
          try {
            addUseOnt((Use_Ontology)obj);
          } catch (DupeKeyException x) {
	    perr = new SHOEException(SHOEException.DUPLICATE_PREFIX,
				     obj.getKey());
	    parseErrs.addElement(perr.getMessage());
	  }
        }
	else if (tagname.equals("DEF-CATEGORY")) {
	  obj = new Def_Category(sgmltag);
	  try {
	    addElement(obj);
	  } catch (DupeKeyException x) {
	    perr = new SHOEException(SHOEException.DUPLICATE_NAME,
				     sgmltag.toString());
	    parseErrs.addElement(perr.getMessage());
	  }
	}
	else if (tagname.equals("DEF-RELATION")) {
	  obj = new Def_Relation(sgmltag);
	  subErrs = obj.readSubTags(intokens);
	  SHOEException.addSubErrors(parseErrs, 
				     "In Def-Relation " + obj.getKey(),
				     subErrs);
	  try {
	    addElement(obj);
	  } catch (DupeKeyException x) {
	    perr = new SHOEException(SHOEException.DUPLICATE_NAME,
				     sgmltag.toString());
	    parseErrs.addElement(perr.getMessage());
	  }		
	}
	else if (tagname.equals("DEF-RENAME")) {  
 
	  /* check for required fields */
	  if (sgmltag.getAttribute("TO") == null) {
	    perr = new SHOEException(SHOEException.MISSING_FIELD,
				     sgmltag.toString(), "TO");
	    parseErrs.addElement(perr.getMessage());
	  }
	  if (sgmltag.getAttribute("FROM") == null) {
	    perr = new SHOEException(SHOEException.MISSING_FIELD,
				     sgmltag.toString(),"FROM");
	    parseErrs.addElement(perr.getMessage());
	  }
	  obj = new Def_Rename(sgmltag);
	  try {
	    addElement(obj);
	  } catch (DupeKeyException x) {
	    perr = new SHOEException(SHOEException.DUPLICATE_NAME,
				     sgmltag.toString());
	    parseErrs.addElement(perr.getMessage());
	  }
	}
	else if (tagname.equals("DEF-INFERENCE")) {
	  obj = new Def_Inference(sgmltag);
	  subErrs = obj.readSubTags(intokens);
	  SHOEException.addSubErrors(parseErrs,
				     "In Def-Inference " + obj.getKey(),
				     subErrs);
	  addInference((Def_Inference)obj);
	}        
	else if (tagname.equals("DEF-CONSTANT")) {
	  obj = new Def_Constant(sgmltag);
	  try {
	    addElement(obj);
	  } catch (DupeKeyException x) {
	    perr = new SHOEException(SHOEException.DUPLICATE_NAME,
				     sgmltag.toString());
	    parseErrs.addElement(perr.getMessage());
	  }
	}
	else if (tagname.equals("DEF-TYPE")) {
	  obj = new Def_Type(sgmltag);
	  try {
	    addElement(obj);
	  } catch (DupeKeyException x) {
	    perr = new SHOEException(SHOEException.DUPLICATE_NAME,
				     sgmltag.toString());
	    parseErrs.addElement(perr.getMessage());
	  }
	}       
	else {                 // invalid tag
	  perr = new SHOEException(SHOEException.UNEXPECTED_TAG,
				   token.toString());
	  parseErrs.addElement(perr.getMessage());
	}
      }
    } while (ttype != intokens.HT_EOF && endtag == false);
    if (ttype == intokens.HT_EOF) {
      perr = new SHOEException(SHOEException.NO_CLOSING_TAG, 
			       toString(), "</ONTOLOGY>");
      parseErrs.addElement(perr.getMessage());
    }
    return parseErrs;
  }

  /** Checks the ontology for reference errors. These are permanently stored
     in the parseErrs vector. Also initializes the set of top-level
     categories defined in the ontology. */
  public Vector validate(SHOE_Obj parent, OntManager ontman) {

    Enumeration enum;
    SHOE_Obj obj;
    Vector subErrs;

    // System.out.println("Validating ontology " + id + ", v." + version);
    enum = elements.elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Def_Rename) {
	Def_Rename drnm = (Def_Rename)obj;
	subErrs = drnm.validate(this, ontman);
	SHOEException.addSubErrors(parseErrs, "In Def-Rename " + obj.getKey(),
				   subErrs);
      }
      else if (obj instanceof Def_Category) {
	Def_Category dcat = (Def_Category)obj;
	subErrs = dcat.validate(this, ontman);
	SHOEException.addSubErrors(parseErrs, 
				   "In Def-Category " + obj.getKey(),
				   subErrs);
      }
      else if (obj instanceof Def_Constant) {
	subErrs = ((Def_Constant)obj).validate(this, ontman);
	SHOEException.addSubErrors(parseErrs,
				   "In Def-Constant " + obj.getKey(),
				   subErrs);
      }
      else if (obj instanceof Def_Relation) {     // validate argument types
	subErrs = ((Def_Relation)obj).validate(this, ontman);
	SHOEException.addSubErrors(parseErrs,
				   "In Def-Relation " + obj.getKey(),
				   subErrs);
      }
    }

    /* Validate all ontologies used by this ontology */
    Use_Ontology uont;
    Ontology ont;
    String header;
    SHOEException perr;

    enum = use_onts.elements();
    while (enum.hasMoreElements()) {
      uont = (Use_Ontology)enum.nextElement();
      ont = ontman.getOntology(uont);
      if (ont == null) {
	perr = new SHOEException(SHOEException.NO_ONTOLOGY_FILE,
				 uont.toString());
	parseErrs.addElement(perr.getMessage());
      }
      else {
	if (ont.hasParseErrors()) {
	  header = "Found errors in ontology " + ont.getOntLabel() + 
	    " at " + ont.getURL();
	  SHOEException.addSubErrors(parseErrs, header, ont.parseErrs);
	}
      }
    }

    /* Validate inference rules */
    enum = getInferenceRules().elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Def_Inference) {
	// validate relation, category and comparison names
	subErrs = ((Def_Inference)obj).validate(this, ontman);
	SHOEException.addSubErrors(parseErrs,
				   "In Def-Inference " + obj.getKey(),
				   subErrs);
      }
    }

    /* find the set of categories which have local names */
    setLocalNameCats(ontman);

    /* find the set of top level categories */
    setTopLevelCategories(ontman);

    return parseErrs;
  }

  /** Returns true if errors when encountered while parsing the
    ontology. */
  public boolean hasParseErrors() {

    return (!parseErrs.isEmpty());
  }

  /** Retrieves error messages for this ontology and concatenate into a
    single string. */
  public String getParseErrors() {

    StringBuffer concatErrs;
    String estr;
    Enumeration enum;

    concatErrs = new StringBuffer();
    enum = parseErrs.elements();
    while (enum.hasMoreElements()) {
      estr = (String)enum.nextElement();
      concatErrs.append(SHOE_Obj.newline + estr);
    }
    return concatErrs.toString();
  }

  /** Writes the ontology and its components to the given print
    stream in SGML format. */
  public void printSGML(PrintStream out) {
  
    Enumeration enum;
    SHOE_Obj obj;

    out.println(this);
    enum = use_onts.elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      obj.printSGML(out);
    }
    enum = elements.elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      obj.printSGML(out);
    }
    enum = inferences.elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      obj.printSGML(out);
    }
    out.println("</ONTOLOGY>");
  }
 
  /** Returns a string consisting of the corresponding open tag for
     the ontology. */
  public String toString() {
  
    String str;
    
    str = "<ONTOLOGY ID=\"" + id + "\"";
    str = str + " VERSION=\"" + version + "\"";
    if (description != null)
      str = str + " DESCRIPTION=\"" + description + "\"";
    if (back_compatible.isEmpty() == false)
      str = str + " BACKWARD-COMPATIBLE-WITH=\"" +
	getBackCompatString() + "\"";
    if (declarators.isEmpty() == false)
      str = str + " DECLARATORS=\"" + getDeclaratorsString() + "\"";
    str = str + ">";
    return str;
  }

  /** Inserts all Def-Categories and Def-Renames of Def-Categories into
    a hash table which is keyed off of the unique id of the orignal
    categories. This table can be used to determine if a particular
    category has a local name. */
  private void setLocalNameCats(OntManager ontman) {

    Enumeration eltEnum;
    SHOE_Obj    obj;
    Unique_Elt  elmt_id;
    SHOE_Obj    baseObj;

    /* We use the unique id of the represented category as our hash key so
       that we can quickly determine if a category appears in the set. */

    /* NOTE: one disadvantage to this approach is that if the same
       category has been given many aliases in the current ontology,
       only one will appear in the set. */

    /* create a new hashtable, using the default size if there are few
       elements. Note, we cannot automatically create a hashtable with
       capacity equal to number of elements, because this might be 0,
       and creating a hashtable with capacity of 0 is an error. */
    if (elements.size() < 10)
      localNameCats = new Hashtable();
    else
      localNameCats = new Hashtable(elements.size());

    eltEnum = getElements();   
    while (eltEnum.hasMoreElements()) {
      obj = (SHOE_Obj)eltEnum.nextElement();
      if (obj instanceof Def_Category) {
	elmt_id = new Unique_Elt(id, version, obj.getKey());
	localNameCats.put(elmt_id, obj);
      }
      else if (obj instanceof Def_Rename) {
	/* in the case of renames, determine if it references a category */
	/* If the rename is of an object in the current ontology, then
	   we can't use OntManager.getElement() because the ontology
	   is not available in OntManager until after it has been
	   validated. Therefore, we check the rename's home object
	   and get it directly if it is in the current ontology. */
	elmt_id = ((Def_Rename)obj).getFromId();
	if (elmt_id != null) {
	  if (getId().equals(elmt_id.getOntId()) && 
	      getVersion().equals(elmt_id.getOntVer()))
	    baseObj = getElement(elmt_id.getName());
	  else
	    baseObj = ontman.getElement(elmt_id);
	  if (baseObj instanceof Def_Category)
	    localNameCats.put(elmt_id, obj);
	}
      }
    }
  }


  /** Determines which Def-Categories and Def-Renames of Def-Categories
    constitute the most abstract categories defined in the ontology
    and stores them in a list, sorted by name. This set can be used as
    root nodes for trees of the category hierarchy defined by this ontology. */
  private void setTopLevelCategories(OntManager ontman) {

    Unique_Elt   elmt_id;
    Enumeration  candEnum;       // enumeration of candidate objects
    SHOE_Obj     candObj;
    Def_Category candCat;        // candidate Def-Category
    Enumeration  parIds;         // parent category element ids
    boolean      isTop;
    Unique_Elt   nextParId;

    // reinitialize topLevelCats (in case validate has been called earlier)
    topLevelCats = new Vector();

    /* for a member of the candidate set, it is a member of topCats if and
       only if none of its parents are members of the set of categories
       with local names. */
    candEnum = localNameCats.elements();
    while (candEnum.hasMoreElements()) {
      candObj = (SHOE_Obj)candEnum.nextElement();
      if (candObj instanceof Def_Rename) {
	/* If the rename is of an object in the current ontology, then
	   we can't use OntManager.getElement() because the ontology
	   is not available in OntManager until after it has been
	   validated. Therefore, we check the rename's home object
	   and get it directly if it is in the current ontology. */
	elmt_id = ((Def_Rename)candObj).getFromId();
	if (getId().equals(elmt_id.getOntId()) && 
	    getVersion().equals(elmt_id.getOntVer()))
	  candCat = (Def_Category)getElement(elmt_id.getName());
	else
	  candCat = (Def_Category)ontman.getElement(elmt_id);
      }
      else
	candCat = (Def_Category)candObj;
      parIds = candCat.getParentIds();
      isTop = true;
      while (parIds.hasMoreElements() && isTop == true) {
	nextParId = (Unique_Elt)parIds.nextElement();
	if (localNameCats.containsKey(nextParId))
	  isTop = false;
      }
      if (isTop == true)
	topLevelCats.addElement(candObj);
    }
    topLevelCats = SHOE_Obj.sortObjectVector(topLevelCats);
  }

  /** Returns the set of Def-Categories and Def-Renames of Def-Categories
    that constitute the most abstract categories defined in the ontology.
    This set can be used as root nodes for trees of the category
    hierarchy defined by this ontology. This set will be null until
    the validate() method has been executed. Returns an enumeration
    of SHOE_Obj's. */
  public Enumeration getTopLevelCategories() {

    return topLevelCats.elements();
  }

}



