/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

/* 
   Change History:
   09/27/01, JDH: Changed validate() to check that the specified name is
                     actually a relation
   11/03/00, JDH: Used accessor methods to get fields in RelArg.
   10/26/00, JDH: Added accessor methods for name, 
                  Made name and elmt_id private.
		  Added getArgValue() method.
		  Used accessor methods for fields in DefRelation	 
  */

package SHOE;

import java.util.*;
import java.io.*;
import HTML.*;
import java.lang.System;

/** This class is used to describe and manipulate a SHOE Relation. */
public class Relation extends Declaration {

  private String name;
  private Vector relargs;     /* contains arguments of the relation */
  private Unique_Elt elmt_id;   
  SHOE_Obj partOf;

  /** Creates a new relation claim with using the given prefixed name. */
  public Relation(String name) {
		
    this.name = name;
    this.relargs = new Vector();
    contents = new Vector();
  }

  /** Creates a new relation claim from its SGML Tag representation. */
  public Relation(SGML_Tag sgmltag) {
      
    name = sgmltag.getAttribute("NAME");
    relargs = new Vector();
    contents = new Vector();
  }
  

  /** Returns an argument of the relation that has the given position
    number.  Since this scans the set arguments looking for one with
    the correct number, it is not dependent on the order of the arguments.
    If we eventually correct this class to store arguments in sorted
    order, then we can make this function more efficient by using direct
    indicing. */
  public RelArg getArg(int pos) {

    int i;
    RelArg arg;

    for (i=0; i < relargs.size(); i++) {
      arg = (RelArg)relargs.elementAt(i);
      if (arg.getPosition() == pos)
	return arg;
    }
    return null;
  }

  /** Returns the value of an argument of the relation that has the
    given position number. */
  public String getArgValue(int pos) {

    int i;
    RelArg arg;

    for (i=0; i < relargs.size(); i++) {
      arg = (RelArg)relargs.elementAt(i);
      if (arg.getPosition() == pos)
	return arg.getValue();
    }
    return null;
  }


  /** Returns the set of the relation's arguments as an array. */
  public RelArg[] getArgs() {

    int i;
    RelArg arglist[];

    arglist = new RelArg[relargs.size()];
    for (i=0; i < relargs.size(); i++)
      arglist[i] = (RelArg)relargs.elementAt(i);
    return arglist;
  }

  /** Returns the number of arguments in the relation. */
  public final int getArgCount() {

    return relargs.size();
  }

  /** Returns the name of the relation as it appears in its original 
    ontology. If there is no such name, as may be the case if by
    error these is no such ontology, returns an empty string. We do
    not return null, because we want to allow string comparisions on
    the result without doing a test for nulls. */
  public String getBaseName() {

    if (elmt_id == null)
      return "";    
    return elmt_id.getName();
  }


  /** Returns the unique id of the ontology object that defines the
    structure of this relation. */
  public final Unique_Elt getElmtId() {

    return elmt_id;
  }


  public String getKey() {
  
    return name;
  }


  /** Returns the prefixed name of the relation referred to by this claim. */
  public String getName() {

    return name;
  }


  /** Returns a string that formats the relation as a predicate, i.e.,
    relname(arg1, arg2). */
  public String getPredicateText() {

    String str;
    int numarg;
    RelArg arg;
    int blankargs = 0;
    Instance inst;

    inst = (Instance)partOf;
    str = name + "(";
    numarg = relargs.size();
    for (int i=0; i < numarg; i++) {
      arg = (RelArg)relargs.elementAt(i);
      /* we calls equals on the inst key b/c this will always have a
	 value, unlike the argument which may have a null value */
      if (inst.getKey().equals(arg.getValue()))
	str = str + "me";
      /*
      if (arg.getPosition() + blankargs > i + 1) {    // use 'me' for any omitted args
	blankargs = blankargs + 1;
	str = str + "me, ";
      }
      */
      else
	str = str + "\"" + arg.getValue() + "\"";
      if (i < numarg - 1)                   // add comma if not last argument
	str = str + ", ";
    }
    str = str + ")";
    return str;
  }


 /** Returns a phrase suitable for describing the claim given that
    its subject is already known. (i.e., short-para arg2) */
  public String getTextBySubject(OntManager ontman) {

    String str;
    int numarg;
    RelArg arg;
    int blankargs = 0;
    Instance inst;
    Def_Relation drel=(Def_Relation)ontman.getElement(elmt_id);

    inst = (Instance)partOf;
    if(drel != null && drel.getPhrase() != null)
	str = drel.getPhrase() + "   ";
    else 
	str = name + "   ";

    numarg = relargs.size();
    for (int i=1; i < numarg; i++) {
      arg = (RelArg)relargs.elementAt(i);
      /* we calls equals on the inst key b/c this will always have a
	 value, unlike the argument which may have a null value */
      if (inst.getKey().equals(arg.getValue()))
	str = str + "me";
      else
	str = str + "\"" + arg.getValue() + "\"";
      if (i < numarg - 1)                   // add comma if not last argument
	  str = str + ", ";
    }
    return str;
  }    
    

  /** Adds an argumnent to the relation. Same as addArg(arg, true).*/
  public void addArg(RelArg arg) {

    addArg(arg, true);
  }

  /** Add an argument to the relation. The arguments will be stored in 
    sorted order by their POS fields. BUG: It is possible to get the
    arguments out of order with 3-ary or more relations, eg., 3,2,1
    becomes 1,3,2. Is there an efficient way to handle this without
    searching the list everytime an argument is added? */
  public void addArg(RelArg arg, boolean addToContent) {

    int argnum;

    argnum = relargs.size();
    if (arg.getPosition() > argnum || arg.getPosition() < 1)
      relargs.addElement(arg);
    else
      relargs.insertElementAt(arg, arg.getPosition() - 1);

    if (addToContent == true) 
      addSHOEtoContents(arg, SHOE_Obj.newline + "   ");
  }

  /** Insert an argument at the first position in a relation's argument list.
    If addToContent is true, the argument will be inserted after the first
    content token in the corresponding content token vector if that token
    is free text. Otherwise, it will be inserted as the first content token */
  private void insertFirstArg(RelArg newArg, boolean addToContent) {

    int i = 0;
    Object firstContent;
    HTML_Token newtoke;
    RelArg firstArg;

    relargs.insertElementAt(newArg, 0);
    if (addToContent == true) {
      firstContent = contents.elementAt(i);
      if (firstContent instanceof SHOE_Obj)
	i = 0;
      else
	i = 1;
      contents.insertElementAt(newArg, i);
      newtoke = new HTML_Token(HTML_Token.TT_TEXT, SHOE_Obj.newline + "   ");
      contents.insertElementAt(newtoke, i + 1);
    }
    else {
      /* if we are not adding the default argument as content, make sure
	 that the other argument uses the TO/FROM position indicators.
	 It should be an error to use a number when implying arguments,
	 but we will be forgiving here */
      firstArg = (RelArg)relargs.firstElement();
      firstArg.setPosAsText(true);
    }
  }

  /** Sets the value of the argument with the selected position. */
  public void setArgValue(int pos, String value) {

    RelArg arg;

    arg = getArg(pos);
    if (arg != null)
      arg.setValue(value);
    else
      System.out.println("There is no argument for position " + pos);
  }

  /** Sets the prefixed name of the relation. */
  public void setName(String name) {

    this.name = name;
  }

  /** Given an SGML tag representing a SHOE relation, determines if
    it contains the required fields and returns a boolean indicating
    if there are any errors. Specific error messages are appended to
    the errStrings vector. */
  public boolean checkFields(SGML_Tag sgmltag, Vector errStrings) {

    boolean       hasErrors = false;
    SHOEException perr;

    if (sgmltag.getAttribute("NAME") == null) {
      hasErrors = true;
      perr = new SHOEException(SHOEException.MISSING_FIELD,
			       sgmltag.toString(), "NAME");
      errStrings.addElement(perr.getMessage());
    }
    return hasErrors;
  }

	
  public Vector readSubTags(HTML_Tokenizer intokens) {

    boolean endtag = false;
    int ttype;
    HTML_Token token;
    SGML_Tag sgmltag = null;
    String tagname;
    RelArg arg;
    SHOEException perr;
    Vector errStrings = null;

    errStrings = new Vector();
    do {
      ttype = intokens.nextToken();
      token = intokens.getLastToken();
//      System.out.println("Token: " + token);
      if (token instanceof SGML_Tag) { 
	/* skips enclosed whitespace, comments and text */
        sgmltag = (SGML_Tag)token;
//        System.out.println("SGML: " + sgmltag);
        tagname = sgmltag.getName();
        if (tagname.equals("/RELATION")) {
          endtag = true;
	  /* the end tag is not considered part of the contents */
	}
        else if (tagname.equals("ARG")) {
          arg = new RelArg(sgmltag);
	  /* Make sure argument had a correct value for the POS field */
	  if (arg.getPosition() == 0) {
	    perr = new SHOEException(SHOEException.INVALID_ARG_POS, 
				     toString());
	    errStrings.addElement(perr.getMessage());	
	  }  
          addArg(arg, false);
	  contents.addElement(arg);
        }
	else {                 // invalid tag
	  perr = new SHOEException(SHOEException.UNEXPECTED_TAG,
				   token.toString());
	  errStrings.addElement(perr.getMessage());
	  contents.addElement(sgmltag);
	}
      }
      else                                 // text or comment
	contents.addElement(token);
    } while (ttype != intokens.HT_EOF && endtag == false);
    if (ttype == intokens.HT_EOF) {
      perr = new SHOEException(SHOEException.NO_CLOSING_TAG, 
			       toString(), "</RELATION>");
      errStrings.addElement(perr.getMessage());
    }
    return errStrings;
  }


  /** Validates the relation by resolving its name, filling in default
    argument values, and validating the arguments. */
  public Vector validate(SHOE_Obj parent, OntManager ontman) {
    /* is there a better way to access ontman without passing it? */
    /* consider setting the ontman when you create a SHOE_Doc object, then
       all other SHOE objects constructors could be passed the ontman */

    /* Note: When you can't locate the reference object in the ontology,
       the code returns with an error, before default arguments can be
       filled in. Thus, if "TO" is used, the first argument may be missing
       in the resulting Relation. */
    SHOE_Obj obj = null;
    Def_Relation drel = null;
    Instance inst;
    int expargs;
    int numargs;
    RelArg curArg;
    RelArg newArg;
    SHOEException perr;
    Vector errStrings = null;

    errStrings = new Vector();
    /* when the relation is in a Def_Inference, the partOf that is being
       passed is currently the Ontology. Will this work? */
    partOf = parent;

    /* validate relation name */
    try {
      elmt_id = ontman.resolveName(name, parent);
    } catch (SHOEException ex) {
      errStrings.addElement(ex.getMessage());
      return errStrings;
    }

    //    System.out.println("Elmt Id = " + elmt_id);
    
    /** If the relation is part of an inference rule in an ontology,
      getElement() will cause to get caught in an infinite loop 
      reloading the ontology. Therefore, we check if the Relation
      is defined in the current ontology, and if so, get its
      definition directly from the ontology, rather than accessing
      the ontology manager. */
    if (parent instanceof Instance) {
	obj = ontman.getElement(elmt_id);
    }
    else if (parent instanceof Ontology) {
      Ontology ont = (Ontology)parent;
      if (ont.getId().equals(elmt_id.getOntId()) && 
	  ont.getVersion().equals(elmt_id.getOntVer()))
	  obj = ont.getElement(elmt_id.getName());
      else
	  obj = ontman.getElement(elmt_id);
    }

    if (obj == null) {
      perr = new SHOEException(SHOEException.NO_ELEMENT_DEF, name, "?");
      errStrings.addElement(perr.getMessage());
      return errStrings;
    }

    // check that the reference element is a relation
    if (!(obj instanceof Def_Relation)) {
	perr = new SHOEException(SHOEException.NOT_RELATION,
				 elmt_id.getName(),
				 elmt_id.getOntId() + ", v." +
				 elmt_id.getOntVer());
	errStrings.addElement(perr.getMessage());
	return errStrings;
    }
    drel = (Def_Relation)obj;

    /* fill in default arguments (if any) for binary relations */
    expargs = drel.getArity();
    numargs = relargs.size();
    //    System.out.println("Exp=" + expargs + ", Act=" + numargs);
    if (parent instanceof Instance) {
      inst = (Instance)parent;
      if (expargs == 2 && numargs == 1) {         // create default argument
	curArg = (RelArg)relargs.firstElement();
	if (curArg.getPosition() == 1) {
	  newArg = new RelArg(2, inst.getKey());
	  addArg(newArg, true);                 // add format w/ new arg
	}
	else if (curArg.getPosition() == 2) {
	  newArg = new RelArg(1, inst.getKey());
	  insertFirstArg(newArg, false);
	}
	//	System.out.println(relargs);
      }
    }

    /* validate arguments */
    Enumeration enum;
    Vector subErrs;

    enum = relargs.elements();
    while (enum.hasMoreElements()) {
      curArg = (RelArg)enum.nextElement();
      if (curArg.getValue() == null) {               // no VALUE statement
	perr = new SHOEException(SHOEException.MISSING_ARG_VALUE, 
				 toString(), curArg.getKey());
	errStrings.addElement(perr.getMessage());	
      }
      else {
	subErrs = curArg.validate(parent, ontman);
	SHOEException.addSubErrors(errStrings, "In Argument " + 
				   curArg.getPosition(), subErrs);
      }
    }
    return errStrings;
  }

  public void printSGML(PrintStream out) {
  
    Enumeration enum;
    RelArg arg;
    
    
    out.print(this);
    printContents(out);
    out.print("</RELATION>");
  }

  /** Returns the string representation of the tag. Any contents
   or end tags are not included. */
  public String toString() {
  
    String str;
    
    str = "<RELATION NAME=\"" + name + "\"";
    str = str + ">";
    return str;
  }
  
}

