/* SHOE Library
 * Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * Author:  Jeff Heflin
 * Contact: Jeff Heflin (heflin@cs.umd.edu)
 *
 */

/* Change History:
     01/29/01, JDH: Use new version of Instance.getUseOnts()
*/

package SHOE;

import java.lang.*;
import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import HTML.*;

/** A SHOE document. The document can be initialized by parsing an
  input stream or by manually adding instances and ontologies. There
  are methods to write the document back out in HTML form or as
  a summary of the SHOE objects. */
public class SHOE_Doc extends Object {

  Vector pshoe;         // contains SHOE ontologies and instances
  /** list of the tokens and SHOE_Objects immediately inside the document */
  Vector contents; 
  float versionSHOE = 0;            // SHOE Version
  Vector parseErrs;                 // document error strings
  private boolean validated;        // has the validate() method been called?

  /** Creates a new, empty SHOE document. */
  public SHOE_Doc() {

    pshoe = new Vector();
    contents = new Vector();
    parseErrs = new Vector();
    validated = false;
  }

  /** Returns an enumeration of the ontology and/or instance objects
    contained within the document. */
  public Enumeration getPSHOE() {
  
    return pshoe.elements();
  }


  /** Returns all (non-SHOE) HTML tokens associated with the document. */
  /* At some point, we may want this to return an enumeration of all tokens
    that make up the document, but this is hard to do efficiently with the
    current data structures. Anyway, this function is only being used to
    currently find out information about tags other than SHOE that are
    within the document */
  public Enumeration getHTML() {

    // return htkns.elements();

    Vector tokeVect;
    Enumeration enum;
    Object elmt;
    SHOE_Obj obj;
    Enumeration subtokens;
    Object subelmt;

    tokeVect = new Vector();
    enum = contents.elements();
    while (enum.hasMoreElements()) {
      elmt = enum.nextElement();
      if (elmt instanceof HTML_Token)
	tokeVect.addElement(elmt);
      else {
	obj = (SHOE_Obj)elmt;
	subtokens = obj.getContentTokens();
	while (subtokens.hasMoreElements()) {
	  subelmt = subtokens.nextElement();
	  tokeVect.addElement(subelmt);
	}
      }
    }   
    return tokeVect.elements();
  }
  
  /** Returns the SHOE version number used by the document. */
  public float getVersion() {

    return versionSHOE;
  }

  /** Resets the document. */
  public void clear() {

    pshoe = new Vector();
    contents = new Vector();
    parseErrs = new Vector();
    versionSHOE = 0;
    validated = false;
  }  
  
  /** Adds a SHOE instance to the document. */
  public void addInstance(Instance inst) {

    pshoe.addElement(inst);
    addSHOEtoContents(inst);
  }

  /** Adds a SHOE ontology to the document. */
  public void addOntology(Ontology ont) {
  
    pshoe.addElement(ont);
    addSHOEtoContents(ont);
  }

  /** Adds text to the BODY section of the document. The
    new content will be appended before the closing </BODY> tag. */
  /*
  public void addTextToBody(String text) {


  }
  */

  /** Removes a SHOE Object from both the list of parsed SHOE objects and the
     document's content list */
  public void deleteSHOEObject(Object obj) {

    int i;

    /* consider using removeElement() below */
    for (i=0; i < pshoe.size(); i++)
      if (pshoe.elementAt(i) == obj)
	pshoe.removeElementAt(i);
    contents.removeElement(obj);
  }

  /** Given an input stream, creates a vector of SHOE objects 
    (instances and ontologies). Creates an HTML tokenizer using
    the input stream and uses this to pull out each SGML tag.
    Creates pSHOE (a vector of SHOE_obj) and contents (a vector
    of the documents contents) */
  public void parse(InputStream instream) throws IOException {

    HTML_Token token;
    SHOE_Obj obj;
    int ttype;
    SGML_Tag sgmltag;
    HTML_Tokenizer intokens = new HTML_Tokenizer(instream);
    String type;
    String verstr;
    String numstr;
    SHOEException perr;

    ttype = intokens.nextToken();
    while (ttype != intokens.HT_EOF) {
      token = intokens.getLastToken();
      if (ttype == intokens.HT_SGMLTAG) {
        sgmltag = (SGML_Tag)token;

	/* check for SHOE version in META tag */
	if (sgmltag.getName().equals("META")) {
	  type = sgmltag.getAttribute("HTTP-EQUIV");
	  if (type != null && type.equals("SHOE")) {
	    verstr = sgmltag.getAttribute("CONTENT");
	    numstr = verstr.substring(8, verstr.length());
	    try {
	      versionSHOE = Float.valueOf(numstr).floatValue();
	    } catch (Exception NumberFormatException) {
	      System.out.println(verstr + " is not a valid version");
	    }
	    if (versionSHOE != 1.0) {
	      perr = new SHOEException(SHOEException.INCOMPATIBLE_VERSION,
				       Float.toString(versionSHOE));
	      parseErrs.addElement(perr.getMessage());
	      // System.out.println(perr.getMessage());
	    }
	  }
	}
        if (versionSHOE == 1.0) {
	  obj = createSHOEobj(sgmltag, intokens);
	  if (obj != null) {
	    /* SHOE tag, add to both the pshoe (for quick reference) and
	       contents (for ordering and reproduction of the HTML) */
	    pshoe.addElement(obj);   
	    contents.addElement(obj);
	  }
	  else
	    contents.addElement(token);     // ordinary HTML tag
	}
	else
	  contents.addElement(token);     // treat as ordinary tag
      }
      else
        contents.addElement(token);       // comment, declaration or text
      ttype = intokens.nextToken();
    }
  }

  /** Checks the document for SHOE reference errors. These are
    permanently stored in the parseErrs vector. Both instances and
    ontologies are validated. Since any errors are merely added to
    the internal list of errors, it should only be called once per
    document.
    */
  public void validate(OntManager ontman) {

    Enumeration enum;
    SHOE_Obj obj;
    Vector subErrs;
    Enumeration errEnum;

    if (validated == true) {
      System.err.println("SHOE_Doc.validate() can only be called once per document.");
      return;
    }
    enum = pshoe.elements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Instance){
	subErrs = ((Instance)obj).validate(null, ontman);
	if (subErrs.size() > 0) {
	  errEnum = subErrs.elements();	  
	  parseErrs.addElement("Validating Instance " + obj.getKey());
	  while (errEnum.hasMoreElements()) {
	    String errStr = "  " + (String)errEnum.nextElement();
	    parseErrs.addElement(errStr);
	  }
	}
      }
      /* NOTE: added the following code 1/27/99. If Expose starts doing 
	 funky stuff with ontologies, then this could be at fault. */
      else if (obj instanceof Ontology) {
	subErrs = ((Ontology)obj).validate(null, ontman);
	if (subErrs.size() > 0) {
	  errEnum = subErrs.elements();	  
	  parseErrs.addElement("Validating Ontology " +
			       ((Ontology)obj).getOntLabel());
	  while (errEnum.hasMoreElements()) {
	    String errStr = "  " + (String)errEnum.nextElement();
	    parseErrs.addElement(errStr);
	  }
	}
      }
    }
    validated = true;
  }

  /** Returns true if errors were detected when parsing the document. */
  public boolean hasParseErrors() {

    return (!parseErrs.isEmpty());
  }

  /** Retrieves error messages for this document and concatenates them
    into a single string. If the validate() method has been called then
    validation errors will be returned in addition to syntatic errors. 
    */
  public String getParseErrors() {

    StringBuffer concatErrs;
    String estr;
    Enumeration enum;

    concatErrs = new StringBuffer();
    enum = parseErrs.elements();
    while (enum.hasMoreElements()) {
      estr = (String)enum.nextElement();
      concatErrs.append(SHOE_Obj.newline + estr);
    }
    return concatErrs.toString();
  }

  /** Print the HTML text and tokens corresponding to a SHOE Document. */
  public void printHTML(PrintStream out) {
   
    Enumeration enum;
    Object elmt;
    HTML_Token token;
    SHOE_Obj obj;

    enum = contents.elements();
    while (enum.hasMoreElements()) {
      elmt = enum.nextElement();
      if (elmt instanceof HTML_Token) {
	token = (HTML_Token)elmt;
	out.print(token);
      }
      else {
	obj = (SHOE_Obj)elmt;
	obj.printSGML(out);
      }
    }
  }

  /** If the document contains any instances, for each instance prints it's
     information, including use ontologies and claims in the Subject View
     format. */
  public void printSummary(PrintStream out) {

    Enumeration enum;
    Object elmt;
    Instance inst;
    Use_Ontology[] useOnts;
    Enumeration cenum;
    Declaration claim;

    enum = getPSHOE();
    while (enum.hasMoreElements()) {
      elmt = enum.nextElement();
      if (elmt instanceof Instance) {
	inst = (Instance)elmt;
	out.println("Instance: " + inst.getKey());
	out.println("   Use Ontologies:");
	useOnts = inst.getUseOnts();
	for (int i=0; i < useOnts.length; i++) {
	  out.println("      " + useOnts[i].getOntLabel());
	}
	out.println("   Claims:");
	cenum = inst.getDeclarations();
	while (cenum.hasMoreElements()) {
	  claim = (Declaration)cenum.nextElement();
	  out.println("      " + claim.getPredicateText());
	}
      }
    }
  }


  /** Given an SGML tag, returns the corresponding SHOE object
    (this will be a category of SHOE_obj) */
  SHOE_Obj createSHOEobj(SGML_Tag sgmltag, HTML_Tokenizer intokens) {
    
    SHOE_Obj obj;
    Vector subErrs;

    if (sgmltag.getName().equals("INSTANCE")) {
      // obj = new Instance(sgmltag);
      /* Let instance know document so it can update the set of instances. */
      obj = new Instance(sgmltag, this);
      subErrs = obj.readSubTags(intokens);
      SHOEException.addSubErrors(parseErrs, "In Instance " + obj.getKey(),
				 subErrs);
    }
    else if (sgmltag.getName().equals("ONTOLOGY")) {
      obj = new Ontology(sgmltag);
      subErrs = obj.readSubTags(intokens);
      /* Removed the following b/c parse errors are also stored in the
	 ontology, resulting in dupes when validate() is called. To ensure
	 that ontology parse errors, are displayed, we should call validate()
	 if necessary before returning them (but can't w/out an
	 OntManager() object. */
      /* 
      SHOEException.addSubErrors(parseErrs, "In Ontology " + ((Ontology)obj).getOntLabel(),
				 subErrs);
      */
    }
    else
      obj = null;
    return obj;
  }


  /** Inserts the new SHOE element in an appropriate place with the
    document. If the document did not previously have SHOE, the
    SHOE meta tag is added. If the document is empty, a basic
    set of HTML tags are added. */
  private void addSHOEtoContents(SHOE_Obj obj) {

    Object elmt;
    SGML_Tag sgmltag;
    HTML_Token newtoke;
    int i;
    boolean added = false;

    if (contents.size() == 0) {
      createBasicDoc();
    }

    i = 0;
    while (added == false && i < contents.size()) {
      elmt = contents.elementAt(i);
      if (elmt instanceof SGML_Tag) {
	sgmltag = (SGML_Tag)elmt;
	/* if SHOE objects have been added to an ordinary HTML document 
	   then add the SHOE META tag and a newline at the end of the
	   HEAD section */
	if (versionSHOE == 0 && sgmltag.getName().equals("/HEAD")) {
	  newtoke = new HTML_Token(HTML_Token.TT_SGMLTAG,
                    "<META HTTP-EQUIV=\"SHOE\" CONTENT=\"VERSION=1.0\">");
	  contents.insertElementAt(newtoke, i);
	  newtoke = new HTML_Token(HTML_Token.TT_TEXT, SHOE_Obj.newline);
	  contents.insertElementAt(newtoke, i+1);
	  versionSHOE = (float)1.0;
	}
	/* what if another version is specified? */
	
	/* insert SHOE object and a newline at the end of the BODY section */
	if (sgmltag.getName().equals("/BODY")) {  
	  contents.insertElementAt(obj, i);
	  newtoke = new HTML_Token(HTML_Token.TT_TEXT, SHOE_Obj.newline);
	  contents.insertElementAt(newtoke, i+1);
	  added = true;
	}
      }
      i = i + 1;
    }

    /* if document does not contain a </BODY> tag, add the SHOE at the end */
    if (added == false) {           
      contents.addElement(obj);
      newtoke = new HTML_Token(HTML_Token.TT_TEXT, SHOE_Obj.newline);
	  contents.addElement(newtoke);
    }

    /* if document does not contain a </HEAD> tag, add the version to front */
    if (versionSHOE == 0) {
      newtoke = new HTML_Token(HTML_Token.TT_SGMLTAG,
		    "<META HTTP-EQUIV=\"SHOE\" CONTENT=\"VERSION=1.0\">");
      contents.insertElementAt(newtoke, 0);
      newtoke = new HTML_Token(HTML_Token.TT_TEXT, SHOE_Obj.newline);
      contents.insertElementAt(newtoke, 1);
    }
  }

  /** Append an HTML tag (with no attributes) to the contents of
    the document. */
  public void addHtmlTag(String tagname) {

    HTML_Token newtoke;

    newtoke = new SGML_Tag(tagname);
    contents.addElement(newtoke);
  }

  /** Append a string to the contents of the document. */
  public void addString(String text) {

    HTML_Token newtoke;

    newtoke = new HTML_Token(HTML_Token.TT_TEXT, text);
    contents.addElement(newtoke);
  }
    
  /** Append a newline to the contents of the document. */
  public void addNewline() {

    addString(SHOE_Obj.newline);
  }

  /** Initializes the SHOE_Doc with a simple framework for an HTML
    document, including the <HTML>, <HEAD>, and <BODY> tags. */
  private void createBasicDoc() {

    addHtmlTag("HTML");
    addNewline();
    addHtmlTag("HEAD");
    addNewline();
    addHtmlTag("/HEAD");
    addNewline();
    addHtmlTag("BODY");
    addNewline();
    addHtmlTag("/BODY");
    addNewline();
    addHtmlTag("/HTML");
    addNewline();
  }

}


