/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in SHOE_Doc.java
 * and the file License.html for details.
 */

package SHOE;

import java.lang.System;
import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import HTML.*;

/** An abstract class that can be used to refer to any type of SHOE
  tag. In addition to overriding the abstract classes, sub-classes of
  SHOE_Obj must create a constructor that receives an SGML_Tag and
  must have a toString() method. If the object could contain other
  tags, then the class must also have a readSubTags() method. */
public abstract class SHOE_Obj {

  /* define usage constants */

  /** A constant that indicates an argument reprents a variable. */
  public static final byte U_VAR = 1;

  /** A constant that indicates an argument reprents a literal value. */
  public static final byte U_VAL = 2;


  /* define comparison type constants */

  /** A constant that indicates a comparison for equality. */
  public static final int C_EQUAL = 1;

  /** A constant that indicates a comparison for inequality. */
  public static final int C_NOT_EQUAL = 2;

  /** A constant that indicates a comparison that is true if the
    first argument is greater than the second. */
  public static final int C_GREATER = 3;

  /** A constant that indicates a comparison that is true if the
    first argument is greater than or equal to the second. */
  public static final int C_GREATER_OR_EQUAL = 4;

  /** A constant that indicates a comparison that is true if the
    first argument is less than the second. */
  public static final int C_LESS = 5 ;

  /** A constant that indicates a comparison that is true if the
    first argument is less than or equal to the second. */
  public static final int C_LESS_OR_EQUAL = 6;

  /** A constant that indicates that the comparison operator is
    undefined. */
  public static final int C_UNDEFINED = -1;

  /** Used to store child tokens and objects */
  protected Vector contents;  

  /** OS specific line separator */
  protected static String newline;

  // Static initializer
  static {
    // System.out.println("Setting newline static variable");
    newline = System.getProperty("line.separator");
  }

  /** Given a vector of SHOE objects, returns a vector of the same
    objects sorted in order by their key. */
  public static Vector sortObjectVector(Vector shoeVec) {

    Enumeration enum;
    Vector sortElts;

    enum = shoeVec.elements();
    sortElts = sortObjectEnum(enum);
    return sortElts;
  }

  /** Given an enumeration of SHOE objects, returns a vector of the same
    objects sorted in order by their key. */
  public static Vector sortObjectEnum(Enumeration shoeEnum) {

    Vector sortElts;
    SHOE_Obj obj;
    SHOE_Obj indxElt;
    int size;
    int i;
    boolean done;

    sortElts = new Vector();
    while (shoeEnum.hasMoreElements()) {
      obj = (SHOE_Obj)shoeEnum.nextElement();
      size = sortElts.size();
      i = 0;
      done = false;
      while (i < size && done == false) {
	indxElt = (SHOE_Obj)sortElts.elementAt(i);
	if (obj.getKey().compareTo(indxElt.getKey()) < 0) {
	  sortElts.insertElementAt(obj, i);
	  done = true;
	}
	i++;
      }
      if (done == false)
	sortElts.addElement(obj);
    }
    return sortElts;
  }

  /** Returns an identifier for the SHOE object. Depending on the
    type of object, it may or may not be unique. */
  public abstract String getKey();
  
  /** Reads and processes the tokens from an HTML_Tokenizer until 
    the appropriate closing tag is encountered. */
  public abstract Vector readSubTags(HTML_Tokenizer intokens);

  /** Writes the objects and its components to the given print
    stream in SGML format. */
  public abstract void printSGML(PrintStream out);

  /* Return a list of the tokens that make up the SHOE object */
  // public abstract Enumeration getTokens();

  /** Inserts an object into the content list. If the last element in the
     list is a text token, then the new object is inserted before it,
     otherwise it is inserted at the end. This is so that formatting
     is maintained properly since the last element typically sets up
     the position of the closing tag. The pretext argument is a string
     that is inserted before the SHOE tag. */
  protected void addSHOEtoContents(SHOE_Obj obj, String pretext) {

    Object lastelmt;
    int lastindx;
    int i;
    HTML_Token newtoke;

    /* if contents are empty, insert a new line to separate contents
       from the end tag */
    if (contents.isEmpty()) {
      newtoke = new HTML_Token(HTML_Token.TT_TEXT, SHOE_Obj.newline);
      contents.addElement(newtoke);
      i = 0;
    }
    else {
      lastindx = contents.size() - 1;
      lastelmt = contents.elementAt(lastindx);
      if (lastelmt instanceof HTML_Token &&
	  ((HTML_Token)lastelmt).getType() == HTML_Token.TT_TEXT)
	i = lastindx;
      else
	i = lastindx + 1;
    }
    newtoke = new HTML_Token(HTML_Token.TT_TEXT, pretext);
    contents.insertElementAt(newtoke, i);
    contents.insertElementAt(obj, i + 1);
  }

  /** Returns and enumeration specifying all (non-SHOE) tokens contained
     between this SHOE Object's start and end tags. If the object is empty,
     an empty enumeration is returned. */
  protected Enumeration getContentTokens() {

    Vector tokeVect;
    Enumeration enum;
    Object elmt;
    HTML_Token token;
    SHOE_Obj obj;
    Enumeration objTokens;

    tokeVect = new Vector();
    if (contents != null) {
      enum = contents.elements();
      while (enum.hasMoreElements()) {
	elmt = enum.nextElement();
	if (elmt instanceof HTML_Token) {
	  token = (HTML_Token)elmt;
	  tokeVect.addElement(token);
	}
	else {                         // SHOE object
	  obj = (SHOE_Obj)elmt;
	  objTokens = obj.getContentTokens();
	  while (objTokens.hasMoreElements())
	    tokeVect.addElement(objTokens.nextElement());
	}
      }
    }
    return tokeVect.elements();
  }

  /** Print the text, HTML tags and SHOE tags contained within the object. */
  protected void printContents(PrintStream out) {

    Enumeration enum;
    Object elmt;
    HTML_Token token;
    SHOE_Obj obj;

    enum = contents.elements();
    while (enum.hasMoreElements()) {
      elmt = enum.nextElement();
      if (elmt instanceof HTML_Token) {
	token = (HTML_Token)elmt;
	out.print(token);
      }
      else {                         // SHOE object
	obj = (SHOE_Obj)elmt;
	obj.printSGML(out);
      }
    }
  }  

}


